# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from cubical.machines.interval_gain_machine import PerIntervalGains
import numpy as np
from cubical.flagging import FL
import cubical.kernels
from numpy.ma import masked_array

from .abstract_machine import log

class PhaseDiagGains(PerIntervalGains):
    """
    This class implements the diagonal phase-only gain machine.
    """
    def __init__(self, label, data_arr, ndir, nmod, chunk_ts, chunk_fs, chunk_label, options):
        """
        Initialises a diagonal phase-only gain machine.
        
        Args:
            label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            chunk_ts (np.ndarray):
                Times for the data being processed.
            chunk_fs (np.ndarray):
                Frequencies for the data being processsed.
            options (dict): 
                Dictionary of options. 
        """
        PerIntervalGains.__init__(self, label, data_arr, ndir, nmod,
                                  chunk_ts, chunk_fs, chunk_label, options)

        # kernel used in solver is diag-diag in diag mode, else uses full kernel version
        if options.get('diag-data') or options.get('diag-only'):
            self.kernel_solve = cubical.kernels.import_kernel('diag_phase_only')
        else:
            self.kernel_solve = cubical.kernels.import_kernel('phase_only')

        self.phases = self.kernel.allocate_gain_array(self.gain_shape, dtype=self.ftype, zeros=True)
        self.gains = np.empty_like(self.phases, dtype=self.dtype)
        self.gains[:] = np.eye(self.n_cor) 
        self.old_gains = self.gains.copy()

    @classmethod
    def determine_diagonality(cls, options):
        return True

    @classmethod
    def get_full_kernel(cls, options, diag_gains):
        if options.get('diag-data'):
            return cubical.kernels.import_kernel('diag_phase_only')
        else:
            return cubical.kernels.import_kernel('phase_only')

    def get_inverse_gains(self):
        """Returns inverse gains and inverse conjugate gains. For phase-only, conjugation is inverse"""
        gh = self.get_conj_gains()
        return gh, self.gains, 0

    def compute_js(self, obser_arr, model_arr):
        """
        This function computes the J\ :sup:`H`\R term of the GN/LM method. 

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.
            model_arr (np.ndrray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                model visibilities.

        Returns:
            np.ndarray:
                J\ :sup:`H`\R
        """

        gh = self.get_conj_gains()
        jh = self.get_new_jh(model_arr)

        self.kernel_solve.compute_jh(model_arr, self.gains, jh, self.t_int, self.f_int)

        jhr = self.get_new_jhr()

        r = self.get_obs_or_res(obser_arr, model_arr)

        self.kernel_solve.compute_jhr(gh, jh, r, jhr, self.t_int, self.f_int)

        return jhr.imag, self.jhjinv, 0


    @property
    def dof_per_antenna(self):
        """This property returns the number of real degrees of freedom per antenna, per solution interval"""
        return 2

    @staticmethod
    def exportable_solutions():
        """ Returns a dictionary of exportable solutions for this machine type. """

        exportables = PerIntervalGains.exportable_solutions()

        exportables.update({
            "phase": (0., ("dir", "time", "freq", "ant", "corr")),
            "phase.err": (0., ("dir", "time", "freq", "ant", "corr")),
        })

        return exportables

    def importable_solutions(self, grid0):
        """ Returns a dictionary of importable solutions for this machine type. """
        # can import 2x2 complex gain, with a corr1/corr2 axis
        solutions = PerIntervalGains.importable_solutions(self, grid0)

        # but also phase (with a corr axis)
        solutions['phase'] = dict(dir=grid0['dir'] if self.dd_term else [0], ant=grid0['ant'], corr=grid0['corr'], **self.interval_grid)

        return solutions

    def export_solutions(self):
        """ Saves the solutions to a dict of {label: solutions,grids} items. """

        solutions = super(PhaseDiagGains, self).export_solutions()

        # construct phase solutions, applying mask from parent gains
        mask = solutions['gain'][0].mask
        solutions['phase'] = masked_array(self.phases, mask)[..., (0, 1), (0, 1)], self.interval_grid

        # phase error is same as gain error (small angle approx!)
        solutions["phase.err"] = solutions["gain.err"][0][..., (0, 1), (0, 1)], self.interval_grid

        return solutions

    def import_solutions(self, soldict):
        """
        Loads solutions from a dict.

        Args:
            soldict (dict):
                Contains gains solutions which must be loaded.
        """
        # load from phase term, if present
        if "phase" in soldict:
            self.phases[...,(0,1),(0,1)] = soldict["phase"].data
            print("loading phases directly", file=log(0))

        # else try to load from gain term
        elif "gain" in soldict:
            self.phases[..., (0,1), (0,1)] = np.angle(soldict["gain"].data)[..., (0,1), (0,1)]
            self.phases[...,0,1].fill(0)
            self.phases[...,1,0].fill(0)
            print("loading phase component from gain", file=log(0))
        else:
            return

        # loaded -- do the housekeeping

        self.restrict_solution(self.phases)
        np.multiply(self.phases, 1j, out=self.gains)
        np.exp(self.gains, out=self.gains)
        self._gains_loaded = True


    def implement_update(self, jhr, jhjinv):

        # variance of gain is diagonal of jhjinv
        if self.posterior_gain_error is None:
            self.posterior_gain_error = np.sqrt(jhjinv.real)
        else:
            np.sqrt(jhjinv.real, out=self.posterior_gain_error)

        update = self.init_update(jhr)

        self.kernel_solve.compute_update(jhr, jhjinv, update)

        if self.iters%2 == 0:
            update *= 0.5
        self.phases += update

        self.restrict_solution(self.phases)

        np.multiply(self.phases, 1j, out=self.gains)
        np.exp(self.gains, out=self.gains)
        self.gains[...,0,1].fill(0)
        self.gains[...,1,0].fill(0)

        # raise flag so updates of G^H and G^-1 are computed
        self._gh_update = self._ghinv_update = True

    def restrict_solution(self, phases):
        """
        Restricts the solution by invoking the inherited restrict_soultion method and applying
        any machine specific restrictions.
        """

        # PerIntervalGains.restrict_solution(self)

        if self.ref_ant is not None:
            phases -= phases[:,:,:,self.ref_ant,0,0][:,:,:,np.newaxis,np.newaxis,np.newaxis]
        for idir in self.fix_directions:
            phases[idir, ...] = 0


    def precompute_attributes(self, data_arr, model_arr, flags_arr, noise):
        """
        Precompute (J\ :sup:`H`\J)\ :sup:`-1`, which does not vary with iteration.

        Args:
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities.
        """
        PerIntervalGains.precompute_attributes(self, data_arr, model_arr, flags_arr, noise)

        self.jhjinv = np.zeros_like(self.gains)

        self.kernel_solve.compute_jhj(model_arr, self.jhjinv, self.t_int, self.f_int)

        self.kernel_solve.compute_jhjinv(self.jhjinv, self.jhjinv, self.gflags, self.eps, FL.ILLCOND)

        self.jhjinv = self.jhjinv.real




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details

# This module has been adapted from the DDFacet package,
# (c) Cyril Tasse et al., see http://github.com/saopicc/DDFacet

import configparser
from collections import OrderedDict
import re

def test():
    P=Parset()

### dead code?
# def FormatDico (DicoIn):
#     Dico=OrderedDict()
#     for key in DicoIn.keys():
#         Dico[key] = ParseConfigString(DicoIn[key])
#     return Dico


def parse_as_python(string, allow_builtins=False, allow_types=False):
    """Tries to interpret string as a Python object. Returns value, or string itself if unsuccessful.
    Names of built-in functions are _not_ interpreted as functions!
    """
    try:
        value = eval(string, {}, {})
        if type(value) is type(all) and not allow_builtins:  # do not interpret built-in function names
            return string
        if type(value) is type(int) and not allow_types:  # do not interpret built-in function names
            return string
        return value
    except:
        return string

def parse_config_string(string, name='config', extended=True, type=None):
    """
    Parses configuration string, converting it to a Python object (i.e. bool, int, float, etc.).
    Can also create a dict of attributes formed as as described below.

    Config string can contain an optional embedded comment, preceded by a '#' symbol. This
    is converted into a docstring, and returned as attrs['doc']. Docstrings can contain optional
    #ATTRIBUTE:VALUE entries. These are extracted and removed from the docstring, and returned
    in the attrs dict.

    The 'type' attribute has special meaning. If present, it forces the string to be parsed
    as a specific type. This overrides the 'type' argument.

    If extended=False, then it DOESN'T attempt to parse out the docstring and attributes,
    but simply tries to interpret the object as the given type. This mode is suitable to
    parsing command-line arguments.

    Args:
        string:     string value to parse
        name:       name of option (used for error messages)
        extended:   if True, enables parsing of docstring and attributes
        type:       forces string to be interpreted as a specific type

    Returns:
        tuple of value, attribute_dict
    """
    if string is None:
        return None, {}

    attrs = {}
    if extended:
        # parse out docstring
        if "#" in string:
            string, docstring = string.split("#", 1)
            string = string.strip()
            # parse out attributes
            while True:
                docstring = docstring.strip()
                # find instance of #attr:value in docstring
                match = re.match("(.*)#(\w+):([^\s]*)(.*)", docstring, re.DOTALL)
                if not match:
                    break
                # extract attribute
                attrname, value = match.group(2), match.group(3)
                # #options:value, value is always treated as a string. Otherwise, treat as Python expression
                if attrname not in ("options",):
                    value = parse_as_python(value)
                attrs[attrname] = value
                # remove it from docstring
                docstring = match.group(1) + match.group(4)
        else:
            docstring = ""
        attrs["doc"] = docstring

        # if attributes contain a type, parse this out
        if 'type' in attrs:
            if attrs['type'] == 'string':
                attrs['type'] = str
            # type had better be a callable type object
            attrs['type'] = type = parse_as_python(attrs['type'], allow_types=True)
            if not callable(type):
                raise ValueError("%s: invalid '#type:%s' attribute"%(name, attrs['type']))

    # if attributes contain an option list, enforce this
    if 'options' in attrs:
        attrs['options'] = opts = attrs['options'].split("|")
        if string not in opts:
            raise ValueError("%s: value %s not in options list"%(name, string))

    # make sure _Help is interpreted as a string
    if name == "_Help":
        return string, attrs

    # interpret explicit types
    if type:
        # make sure None string is still None
        if type is str and string == "None" or string == "none":
            return None, attrs
        # make sure False/True etc. are interpreted as booleans
        if type is bool:
            return bool(parse_as_python(string)), attrs
        return type(string), attrs

    # Now, some kludges for backward compatibility
    # A,B,C and [A,B,C] are parsed to a list
    as_list = len(string)>1 and string[0] == '[' and string[-1] == ']'
    if as_list:
        string = string[1:-1]
    if as_list or "," in string:
        return [ parse_as_python(x) for x in string.split(",") ], attrs

    # Otherwise just interpret the value as a Python object if possible
    return parse_as_python(string), attrs


class Parset():
    def __init__(self, filename=None):
        """Creates parset, reads from file if specified"""
        self.value_dict = self.DicoPars = OrderedDict()   # call it DicoPars for compatibility with old testing code
        self.attr_dict = OrderedDict()
        if filename:
            self.read(filename)

    def update_values (self, other, other_filename=''):
        """Updates this Parset with keys found in other parset.
        other_filename is only needed for error messages."""
        for secname, secvalues in other.value_dict.items():
            if secname in self.value_dict:
                for name, value in secvalues.items():
                    attrs = self.attr_dict[secname].get(name)
                    if attrs is None:
                        attrs = self.attr_dict[secname][name] = \
                            other.attr_dict[secname].get(name, {})
                    if attrs.get('cmdline_only'):
                        continue
                    # check value for type and options conformance
                    if 'type' in attrs:
                        try:
                            value = attrs['type'](value)
                        except:
                            raise TypeError("invalid [{}] {}={} setting{}".format(
                                            secname, name, value, other_filename))
                    if 'options' in attrs and value not in attrs['options']:
                        if str(value) in attrs['options']:
                            value = str(value)
                        else:
                            raise TypeError("invalid [{}] {}={} setting{}".format(
                                            secname, name, value, other_filename))
                    self.value_dict[secname][name] = value
                    # make sure aliases get copied under both names
                    alias = attrs.get('alias') or attrs.get('alias_of')
                    if alias:
                        self.value_dict[secname][alias] = value
            else:
                self.value_dict[secname] = secvalues
                self.attr_dict[secname] = other.attr_dict[secname]

    def read (self, filename, default_parset=False):
        """Reads parset from filename.
        default_parset: if True, this is treated as the default parset, and things like templated
        section names are expanded.
        """
        self.filename = filename
        self.Config = config = configparser.ConfigParser(dict_type=OrderedDict)
        config.optionxform = str
        success = config.read(self.filename)
        self.success = bool(len(success))
        if self.success:
            self.sections = config.sections()
            for section in self.sections:
                self.value_dict[section], self.attr_dict[section] = self.read_section(config, section)
        # # now migrate from previous versions
        # self.version = self.value_dict.get('Misc', {}).get('ParsetVersion', 0.0)
        # if self.version != 0.1:
        #     self._migrate_ancient_0_1()
        #     self.migrated = self.version
        #     self.version = self.value_dict['Misc']['ParsetVersion'] = 0.1
        # else:
        #     self.migrated = None
        # # if "Mode" not in self.value_dict["Output"] and "Mode" in self.value_dict["Image"]:
        # #     self.value_dict["Output"]["Mode"] = self.value_dict["Image"]["Mode"]
        # #     del self.value_dict["Image"]["Mode"]

    def read_section(self, config, section):
        """Returns two dicts corresponding to the given section: a dict of option:value,
        and a dict of option:attribute_dict"""
        dict_values = OrderedDict()
        dict_attrs = OrderedDict()
        for option in config.options(section):
            strval = config.get(section, option)
            # option names with an "|" in them specify a longhand alias
            if "|" in option:
                option, alias = option.split("|",1)
            else:
                alias = None
            dict_values[option], dict_attrs[option] = parse_config_string(strval, name=option)
            # if option has an alias, mke copy of both in dicts
            if alias:
                dict_attrs[option]['alias'] = alias
                dict_values[alias] = dict_values[option]
                dict_attrs[alias] = { 'alias_of': option }
        return dict_values, dict_attrs

    def set (self, section, option, value):
        self.value_dict.setdefault(section,{})[option] = value

    def write (self, f):
        """Writes the Parset out to a file object"""
        for section, content in self.value_dict.items():
            f.write('[%s]\n'%section)
            for option, value in content.items():
                attrs = self.attr_dict.get(section, {}).get(option, {})
                if option[0] != "_" and not attrs.get('cmdline_only') and not attrs.get('alias_of'):
                    f.write('%s = %s \n'%(option, str(value)))
            f.write('\n')

    def _makeSection (self, section):
        """
        Helper method for migration: makes a new section
        """
        for dd in self.value_dict, self.attr_dict:
            dd.setdefault(section, OrderedDict())
        return section

    def _renameSection (self, oldname, newname):
        """
        Helper method for migration: renames a section. If the new section already exists, merges options into
        it.
        """
        for dd in self.value_dict, self.attr_dict:
            if oldname in dd:
                dd.setdefault(newname, OrderedDict()).update(dd.pop(oldname))
        return newname

    def _del (self, section, option):
        """
        Helper method for migration: removes an option
        """
        for dd in self.value_dict, self.attr_dict:
            if section in dd and option in dd[section]:
                dd[section].pop(option)

    def _rename (self, section, oldname, newname):
        """
        Helper method for migration: renames an option within a section. Optionally remaps option values using
        the supplied dict.
        """
        for dd in self.value_dict, self.attr_dict:
            if section in dd and oldname in dd[section]:
                dd[section][newname] = dd[section].pop(oldname)

    def _remap (self, section, option, remap):
        """
        Helper method for migration: remaps the values of an option
        """
        if section in self.value_dict and option in self.value_dict[section]:
            value = self.value_dict[section][option]
            if value in remap:
                self.value_dict[section][option] = remap[value]

    def _move (self, oldsection, oldname, newsection, newname):
        """
        Helper method for migration: moves an option to a different section
        """
        for dd in self.value_dict, self.attr_dict:
            if oldsection in dd and oldname in dd[oldsection]:
                dd.setdefault(newsection, OrderedDict())[newname] = dd[oldsection].pop(oldname)




import numpy as np
from builtins import range

# these control the layout of saved plots
DPI = 150.  # resolution: determines size of text relative to plots
ZOOM = 10  # image zoom. Roughly determines size of "pixels" when rendering images

def make_antenna_xaxis(antnames):
    import pylab
    """Helper function. Sets up a plot X axis that shows antenna numbers nicely"""
    nant = len(antnames)
    # make a tick at every antenna, but label only the originally labelled ones
    locs = np.arange(nant)
    labels = [''] * nant
    pylab.xlim(-1, nant)
    for x, lbl in zip(*pylab.xticks()):
        x = int(x)
        if x >= 0 and x < nant:
            labels[x] = antnames[x]
    pylab.xticks(locs, labels)
    # draw vertical guides for antennas -- thicker line every 5 antennas
    for x in range(0, nant):
        pylab.axvline(x, c="grey", lw=0.5 if x%5 else 1, ls=':' if x%5 else '-')
    pylab.xlabel("antenna")


def make_summary_plots(st, ms, GD, basename):
    """

    Args:
        st:         statistics.SolverStats object
        GD:         dict of global options
        basename:   base filename for plots
    Returns:

    """
    stats.make_stats_plots(st, GD, basename, ms.metadata)

from . import ifrgains, stats




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from builtins import range
from cubical.database.pickled_db import PickledDatabase
from cubical.data_handler.ms_data_handler import MSDataHandler
from cubical.tools import logger
from pyrap.tables import table as tbl
import os
import shutil
import numpy as np
import subprocess
import six

log = logger.getLogger("casa_db_adaptor")

# to the tune of KATDAL :)
BLANK_TABLE_NAME = 'blankcaltable.CASA'
BLANK_TABLE_TARBALL = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                    "{}.tgz".format(BLANK_TABLE_NAME))

class casa_caltable_factory(object):
        """
            Gaintable Factory
            
            Useful methods for creating and writing CASA-style gaintables
        """
        
        @classmethod
        def init_empty(cls, db, filename, solfreqs, solants, field_ndir=1, is_complex=True, viscal_label="B Jones"):
            """
                Initialize empty calibration table with metadata
                
                Args:
                    db: a cubical pickled_db instance
                    filename: filename of CASA gaintable to be written
                    solfreqs: centre frequencies of solutions to be stored in table
                    solants: Names from Cubical database antenna axis, used to order solutions
                    field_ndir: Number of solvable directions in field
                    is_complex: Solutions are complex or real-valued
                    viscal_label: Sets viscal property of CASA table - used as identifier in CASA
            """
            if six.PY3:
                log.error("Gaintables cannot be written in Python 3 mode due to current casacore implementation issues")
                return
            if os.path.exists(filename):
                if os.path.isfile(filename):
                    log.error("CASA calibration table destination already exists but is not a directory. Will not remove.")    
                    return
                else:
                    log.info("Destination CASA gain table '%s' exists. Will overwrite." % filename)
                    shutil.rmtree(filename) # CASA convention is to overwrite
                    
            basedir = os.path.dirname(filename) or '.'
            subprocess.check_output(["tar", "zxvf", BLANK_TABLE_TARBALL, "-C", basedir])
            os.rename(os.path.join(basedir, BLANK_TABLE_NAME), filename)

            antorder = [db.antnames.index(an) for an in solants]
            with tbl("%s::ANTENNA" % str(filename), ack=False, readonly=False) as t:
                t.addrows(nrows=len(db.anttype))
                t.putcol("OFFSET", db.antoffset[antorder])
                t.putcol("POSITION", db.antpos[antorder])
                t.putcol("TYPE", np.array(db.anttype)[antorder])
                t.putcol("DISH_DIAMETER", db.antdishdiam[antorder])
                t.putcol("FLAG_ROW", db.antflagrow[antorder])
                t.putcol("MOUNT", np.array(db.antmount)[antorder])
                t.putcol("NAME", np.array(db.antnames)[antorder])
                t.putcol("STATION", np.array(db.antstation)[antorder])
                
            assert "field" in db.metadata, "Solver field not passed in metadata. This is a bug"
            assert type(db.metadata["field"]) is int, "Currently only supports single field"
            selfield = np.arange(len(db.fieldname)) == db.metadata["field"]
            with tbl("%s::FIELD" % str(filename), ack=False, readonly=False) as t:
                t.addrows(nrows=field_ndir)
                t.putcol("DELAY_DIR", np.tile(db.fielddelaydirs[selfield], (field_ndir, 1)))
                t.putcol("PHASE_DIR", np.tile(db.fieldphasedirs[selfield], (field_ndir, 1)))
                t.putcol("REFERENCE_DIR", np.tile(db.fieldrefdir[selfield], (field_ndir, 1)))
                t.putcol("CODE", np.tile(np.array(db.fieldcode)[selfield], (field_ndir, 1)))
                t.putcol("FLAG_ROW", np.tile(db.fieldflagrow[selfield], (field_ndir, 1)))
                t.putcol("NAME", np.array(map(str, ["%s_DIR_%d" % (f, fdi) for fdi, f in enumerate([db.fieldname[np.where(selfield)[0][0]]] * field_ndir)])).T)
                t.putcol("SOURCE_ID", np.tile(db.fieldsrcid[selfield], (field_ndir, 1)) + np.arange(field_ndir).T)
                t.putcol("TIME", np.tile(db.fieldtime[selfield], (field_ndir, 1)))
    
            with tbl("%s::OBSERVATION" % str(filename), ack=False, readonly=False) as t:
                t.addrows(nrows=len(db.obsobserver))
                (len(db.obstimerange) != 0) and t.putcol("TIME_RANGE", db.obstimerange)
                (len(db.obslog) != 0) and t.putcol("LOG", db.obslog)
                (len(db.obsschedule) != 0) and t.putcol("SCHEDULE", db.obsschedule)
                (len(db.obsflagrow) != 0) and t.putcol("FLAG_ROW", db.obsflagrow)
                (len(db.obsobserver) != 0) and t.putcol("OBSERVER", db.obsobserver)
                (len(db.obsproject) != 0) and t.putcol("PROJECT", db.obsproject)
                (len(db.obsreleasedate) != 0) and t.putcol("RELEASE_DATE", db.obsreleasedate)
                (len(db.obstelescopename) != 0) and t.putcol("TELESCOPE_NAME", db.obstelescopename)
            
            with tbl("%s::SPECTRAL_WINDOW" % str(filename), ack=False, readonly=False) as t:
                t.addrows(nrows=len(db.sel_ddids))
                # Per DDID determine solution spacing in frequency
                for iddid, ddid in enumerate(db.sel_ddids):
                    spwid = db.ddid_spw_map[ddid]
                    minfreq = np.min(db.spwchanfreq[spwid] - 0.5 * db.spwchanwidth[spwid])
                    maxfreq = np.max(db.spwchanfreq[spwid] + 0.5 * db.spwchanwidth[spwid]) 
                    ddsolfreqs = solfreqs[np.logical_and(solfreqs >= minfreq,
                                                         solfreqs <= maxfreq)]
                    # assume linearly-spaced solutions for N - 1 solutions along grid and last solution 
                    # may have different spacing if solution interval does not exactly divide number of channels
                    if ddsolfreqs.size > 1:
                        ch0 = np.min(db.spwchanfreq)
                        chN = np.max(db.spwchanfreq)
                        ddsolwidthmax = 2 * (ddsolfreqs[0] - ch0)
                        ddsolwidth = np.ones(ddsolfreqs.size) * ddsolwidthmax
                        ddsolwidth[-1] = ddsolwidthmax + (chN - (ddsolfreqs.size * ddsolwidthmax + ch0))
                    else:
                        ddsolwidth = db.spwtotalbandwidth
                        
                    t.putcell("MEAS_FREQ_REF", iddid, db.spwmeasfreq[spwid])
                    t.putcell("CHAN_FREQ", iddid, ddsolfreqs)
                    t.putcell("REF_FREQUENCY", iddid, db.spwreffreq[spwid])
                    t.putcell("CHAN_WIDTH", iddid, ddsolwidth)
                    t.putcell("EFFECTIVE_BW", iddid, ddsolwidth) # TODO: this may not be true
                    t.putcell("RESOLUTION", iddid, db.spwresolution[spwid])
                    t.putcell("FLAG_ROW", iddid, db.spwflagrow[spwid])
                    t.putcell("FREQ_GROUP", iddid, db.spwfreqgroup[spwid])
                    t.putcell("FREQ_GROUP_NAME", iddid, db.spwfreqgroupname[spwid])
                    t.putcell("IF_CONV_CHAIN", iddid, db.spwifconvchain[spwid])
                    t.putcell("NAME", iddid, str(db.spwname[spwid]))
                    t.putcell("NET_SIDEBAND", iddid, db.spwnetsideband[spwid])
                    t.putcell("NUM_CHAN", iddid, ddsolfreqs.size)
                    t.putcell("TOTAL_BANDWIDTH", iddid, maxfreq - minfreq)
                    
            with tbl(str(filename), ack=False, readonly=False) as t:
                t.putkeyword("ParType", "Complex" if is_complex else "Float")
                t.putkeyword("VisCal", viscal_label)
                
                if not is_complex:
                    cdesc = t.getcoldesc("CPARAM")
                    cdesc["valueType"] = "float"
                    t.addcols({"FPARAM": cdesc})
                    t.removecols("CPARAM")
                        
        @classmethod
        def create_G_table(cls, db, gname, outname = "Gphase"):
            """
                Write diagonal G-phase caltable
                
                Args:
                    db: a cubical pickled_db instance
                    gname: name of pickled_db solutions to export
                    outname: suffix of exported CASA gaintable
            """
            if six.PY3:
                log.error("Gaintables cannot be written in Python 3 mode due to current casacore implementation issues")
                return
            if np.prod(db[gname].shape) == 0:
                log.warn("No %s solutions. Will not write CASA table" % gname)
                return
            paramerrs = cls.__check_param_err(db, gname)
            assert db[gname].axis_labels == ('dir', 'time', 'freq', 'ant', 'corr'), "DB table in unrecognized format"
            
            ddids = db.sel_ddids
            ndir = len(db[gname].grid[db[gname].ax.dir])
            ntime = len(db[gname].grid[db[gname].ax.time])
            nant = len(db[gname].grid[db[gname].ax.ant])
            ncorr = len(db[gname].grid[db[gname].ax.corr])
            nddids = len(db.sel_ddids)
            nrow = ndir * ntime * \
                    nant * len(ddids)
            assert ncorr == 2, "Expected diagnonal Jones matrix"
            
            cls.init_empty(db, 
                           db.filename + ".%s.casa" % outname, 
                           db[gname].grid[db[gname].ax.freq],
                           db[gname].grid[db[gname].ax.ant],
                           field_ndir=ndir,
                           viscal_label="G Jones")
            
            with tbl(str(db.filename + ".%s.casa" % outname), ack=False, readonly=False) as t:
                t.addrows(nrows=nrow)
                for iddid, ddid in enumerate(db.sel_ddids):
                    spwid = db.ddid_spw_map[ddid]
                    minfreq = np.min(db.spwchanfreq[spwid] - 0.5 * db.spwchanwidth[spwid])
                    maxfreq = np.max(db.spwchanfreq[spwid] + 0.5 * db.spwchanwidth[spwid]) 
                    ddsolfreqindx = np.argwhere(np.logical_and(db[gname].grid[db[gname].ax.freq] >= minfreq,
                                                               db[gname].grid[db[gname].ax.freq] <= maxfreq))
                    params = np.swapaxes(db[gname].get_cube()[:, :, ddsolfreqindx, :, :],
                                         2, 3).reshape(ndir * ntime * nant, len(ddsolfreqindx), ncorr) 
                    paramerrs = np.swapaxes(paramerrs[:, :, ddsolfreqindx, :, :],
                                            2, 3).reshape(ndir * ntime * nant, len(ddsolfreqindx), ncorr) 
                    flags = np.ma.getmaskarray(params)
                    fieldid = np.repeat(np.arange(ndir), ntime * nant) # dir (marked as field) is slowest varying
                    time = np.repeat(np.tile(db[gname].grid[db[gname].ax.time], ndir), nant)
                    ant1 = np.tile(np.arange(nant), ndir * ntime) # FK ndir * ntime blocks
                    #ant2 can be the same - it is not used unless specifying mueller matricies
                    nrowsdd = ntime *  nant * ndir
                    t.putcol("TIME", time, startrow=nrowsdd * iddid)
                    t.putcol("FIELD_ID", fieldid, startrow=nrowsdd * iddid)
                    t.putcol("SPECTRAL_WINDOW_ID", np.ones(ant1.shape) * iddid, startrow=nrowsdd * iddid) # new spectral window with freq range for these sols
                    t.putcol("ANTENNA1", ant1, startrow=nrowsdd * iddid)
                    t.putcol("ANTENNA2", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) 
                    t.putcol("INTERVAL", np.zeros(ant1.shape), startrow=nrowsdd * iddid) #TODO: unclear from MEMO 229
                    t.putcol("SCAN_NUMBER", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) #TODO this FK info is not available yet @oms
                    t.putcol("OBSERVATION_ID", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) #TODO this FK info is not available yet @oms
                    t.putcol("CPARAM", np.ma.getdata(params), startrow=nrowsdd * iddid)
                    t.putcol("PARAMERR", np.ma.getdata(paramerrs), startrow=nrowsdd * iddid)
                    t.putcol("FLAG", flags, startrow=nrowsdd * iddid)
                    t.putcol("SNR", np.ones(params.shape) * np.inf, startrow=nrowsdd * iddid) #TODO this is not available @oms
                    t.putcol("WEIGHT", np.ones(params.shape), startrow=nrowsdd * iddid) #TODO this is not available @oms
        @classmethod
        def __check_param_err(cls, db, gname):
            if db[gname].shape != db[gname + ".err"].shape:
                log.warn("PARAM err shape %s does not match PARAM shape %s. This is a solver bug so we will assume infinite errors on all solutions." %
                         (",".join(map(str, db[gname].shape)),
                          ",".join(map(str, db[gname + ".err"].shape))))
                return np.zeros_like(db[gname].get_cube(), dtype=db[gname + ".err"].get_cube().dtype) * np.inf
            return db[gname + ".err"].get_cube()
        @classmethod
        def create_B_table(cls, db, gname, outname = "B", diag=True):
            """
                Write diagonal B-Jones caltable
                
                Args:
                    db: a cubical pickled_db instance
                    gname: name of pickled_db solutions to export
                    outname: suffix of exported CASA gaintable
                    diag: Write out diagonal of Jones matrix if true, off-diagonal (leakage) terms otherwise.
            """
            if six.PY3:
                log.error("Gaintables cannot be written in Python 3 mode due to current casacore implementation issues")
                return
            if np.prod(db[gname].shape) == 0:
                log.warn("No %s solutions. Will not write CASA table" % gname)
                return
            paramerrs = cls.__check_param_err(db, gname)
            assert db[gname].axis_labels == ('dir', 'time', 'freq', 'ant', 'corr1', 'corr2'), "DB table in unrecognized format"
            
            ddids = db.sel_ddids
            ndir = len(db[gname].grid[db[gname].ax.dir])
            ntime = len(db[gname].grid[db[gname].ax.time])
            nant = len(db[gname].grid[db[gname].ax.ant])
            ncorr1 = len(db[gname].grid[db[gname].ax.corr1])
            ncorr2 = len(db[gname].grid[db[gname].ax.corr2])
            nddids = len(db.sel_ddids)
            nrow = ndir * ntime * \
                    nant * len(ddids)
            assert ncorr1 == ncorr2 and ncorr1 == 2, "Expected 2x2 solution, this is a bug"
            
            cls.init_empty(db, 
                           db.filename + ".%s.casa" % outname, 
                           db[gname].grid[db[gname].ax.freq],
                           db[gname].grid[db[gname].ax.ant],
                           field_ndir=ndir,
                           viscal_label="B Jones" if diag else "D Jones")
            
            with tbl(str(db.filename) + ".%s.casa" % outname, ack=False, readonly=False) as t:
                t.addrows(nrows=nrow)
                
                for iddid, ddid in enumerate(db.sel_ddids):
                    spwid = db.ddid_spw_map[ddid]
                    minfreq = np.min(db.spwchanfreq[spwid] - 0.5 * db.spwchanwidth[spwid])
                    maxfreq = np.max(db.spwchanfreq[spwid] + 0.5 * db.spwchanwidth[spwid]) 
                    ddsolfreqindx = np.argwhere(np.logical_and(db[gname].grid[db[gname].ax.freq] >= minfreq,
                                                               db[gname].grid[db[gname].ax.freq] <= maxfreq))
                    jones_entries = [0, 3] if diag else [1, 2] # diagonal or crosshands
                    params = np.swapaxes(db[gname].get_cube()[:, :, ddsolfreqindx, :, :], 
                                         2, 3).reshape(ndir * ntime * nant, len(ddsolfreqindx), ncorr1 * ncorr2)[:, :, jones_entries]
                    paramerrs = np.swapaxes(paramerrs[:, :, ddsolfreqindx, :, :],
                                            2, 3).reshape(ndir * ntime * nant, len(ddsolfreqindx), ncorr1 * ncorr2)[:, :, jones_entries]
                    flags = np.ma.getmaskarray(params)
                    fieldid = np.repeat(np.arange(ndir), ntime * nant) # dir (marked as field) is slowest varying
                    time = np.repeat(np.tile(db[gname].grid[db[gname].ax.time], ndir), nant)
                    ant1 = np.tile(np.arange(nant), ndir * ntime) # FK ndir * ntime blocks
                    #ant2 can be the same - it is not used unless specifying mueller matricies
                    nrowsdd = ntime *  nant * ndir
                    t.putcol("TIME", time, startrow=nrowsdd * iddid)
                    t.putcol("FIELD_ID", fieldid, startrow=nrowsdd * iddid)
                    t.putcol("SPECTRAL_WINDOW_ID", np.ones(ant1.shape) * iddid, startrow=nrowsdd * iddid) # new spectral window with freq range for these sols
                    t.putcol("ANTENNA1", ant1, startrow=nrowsdd * iddid)
                    t.putcol("ANTENNA2", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) 
                    t.putcol("INTERVAL", np.zeros(ant1.shape), startrow=nrowsdd * iddid) #TODO: unclear from MEMO 229
                    t.putcol("SCAN_NUMBER", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) #TODO this FK info is not available yet @oms
                    t.putcol("OBSERVATION_ID", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) #TODO this FK info is not available yet @oms
                    t.putcol("CPARAM", np.ma.getdata(params), startrow=nrowsdd * iddid)
                    t.putcol("PARAMERR", np.ma.getdata(paramerrs), startrow=nrowsdd * iddid)
                    t.putcol("FLAG", flags, startrow=nrowsdd * iddid)
                    t.putcol("SNR", np.ones(params.shape) * np.inf, startrow=nrowsdd * iddid) #TODO this is not available @oms
                    t.putcol("WEIGHT", np.ones(params.shape), startrow=nrowsdd * iddid) #TODO this is not available @oms
                    
        @classmethod
        def create_D_table(cls, db, gname, outname = "D"):
            """
                Wrapper for off-diagonal B-Jones (leakage / "D") Jones table
                
                Args:
                    db: a cubical pickled_db instance
                    gname: name of pickled_db solutions to export
                    outname: suffix of exported CASA gaintable
            """
            if six.PY3:
                log.error("Gaintables cannot be written in Python 3 mode due to current casacore implementation issues")
                return
            cls.create_B_table(db, gname, outname, diag=False)
        
        @classmethod
        def create_K_table(cls, db, gname, outname = "K"):
            """
                Write real-valued K-Jones table
                
                Args:
                    db: a cubical pickled_db instance
                    gname: name of pickled_db solutions to export
                    outname: suffix of exported CASA gaintable
            """
            if six.PY3:
                log.error("Gaintables cannot be written in Python 3 mode due to current casacore implementation issues")
                return
            if np.prod(db[gname].shape) == 0:
                log.warn("No %s solutions. Will not write CASA table" % gname)
                return
            paramerrs = cls.__check_param_err(db, gname)
            assert db[gname].axis_labels == ('dir', 'time', 'freq', 'ant', 'corr'), "DB table in unrecognized format"
            
            ddids = db.sel_ddids
            ndir = len(db[gname].grid[db[gname].ax.dir])
            ntime = len(db[gname].grid[db[gname].ax.time])
            nant = len(db[gname].grid[db[gname].ax.ant])
            ncorr = len(db[gname].grid[db[gname].ax.corr])
            nddids = len(db.sel_ddids)
            nrow = ndir * ntime * \
                    nant * len(ddids)
            assert ncorr == 2, "Expected diagnonal Jones matrix"
            
            cls.init_empty(db, 
                           db.filename + ".%s.casa" % outname, 
                           db[gname].grid[db[gname].ax.freq],
                           db[gname].grid[db[gname].ax.ant],
                           field_ndir=ndir,
                           is_complex=False,
                           viscal_label="K Jones")
            
            with tbl(str(db.filename + ".%s.casa" % outname), ack=False, readonly=False) as t:
                t.addrows(nrows=nrow)
                for iddid, ddid in enumerate(db.sel_ddids):
                    spwid = db.ddid_spw_map[ddid]
                    minfreq = np.min(db.spwchanfreq[spwid] - 0.5 * db.spwchanwidth[spwid])
                    maxfreq = np.max(db.spwchanfreq[spwid] + 0.5 * db.spwchanwidth[spwid]) 
                    ddsolfreqindx = np.argwhere(np.logical_and(db[gname].grid[db[gname].ax.freq] >= minfreq,
                                                               db[gname].grid[db[gname].ax.freq] <= maxfreq))
                    # note -- CASA K table delays are in nanoseconds. This presumes delays in the cubical tables are already denormalized into seconds
                    params = np.swapaxes(db[gname].get_cube()[:, :, ddsolfreqindx, :, :],
                                         2, 3).reshape(ndir * ntime * nant, len(ddsolfreqindx), ncorr) * 1.0e9
                    paramerrs = np.swapaxes(paramerrs[:, :, ddsolfreqindx, :, :],
                                            2, 3).reshape(ndir * ntime * nant, len(ddsolfreqindx), ncorr) * 1.0e9
                    flags = np.ma.getmaskarray(params)
                    fieldid = np.repeat(np.arange(ndir), ntime * nant) # dir (marked as field) is slowest varying
                    time = np.repeat(np.tile(db[gname].grid[db[gname].ax.time], ndir), nant)
                    ant1 = np.tile(np.arange(nant), ndir * ntime) # FK ndir * ntime blocks
                    #ant2 can be the same - it is not used unless specifying mueller matricies
                    nrowsdd = ntime *  nant * ndir
                    t.putcol("TIME", time, startrow=nrowsdd * iddid)
                    t.putcol("FIELD_ID", fieldid, startrow=nrowsdd * iddid)
                    t.putcol("SPECTRAL_WINDOW_ID", np.ones(ant1.shape) * iddid, startrow=nrowsdd * iddid) # new spectral window with freq range for these sols
                    t.putcol("ANTENNA1", ant1, startrow=nrowsdd * iddid)
                    t.putcol("ANTENNA2", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) 
                    t.putcol("INTERVAL", np.zeros(ant1.shape), startrow=nrowsdd * iddid) #TODO: unclear from MEMO 229
                    t.putcol("SCAN_NUMBER", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) #TODO this FK info is not available yet @oms
                    t.putcol("OBSERVATION_ID", np.ones(ant1.shape) * -1, startrow=nrowsdd * iddid) #TODO this FK info is not available yet @oms
                    t.putcol("FPARAM", np.ma.getdata(params), startrow=nrowsdd * iddid)
                    t.putcol("PARAMERR", np.ma.getdata(paramerrs), startrow=nrowsdd * iddid)
                    t.putcol("FLAG", flags, startrow=nrowsdd * iddid)
                    t.putcol("SNR", np.ones(params.shape) * np.inf, startrow=nrowsdd * iddid) #TODO this is not available @oms
                    t.putcol("WEIGHT", np.ones(params.shape), startrow=nrowsdd * iddid) #TODO this is not available @oms
        
class casa_db_adaptor(PickledDatabase):
    
    def __init__(self):
        """
            CASA-style gaintable adaptor for Cubical pickled_db database writer
            
            This adaptor adds functionality to export databases to CASA gaintables
            which can be read and plotted in CASA plotms or similar diagnostic utilities.
        """
        PickledDatabase.__init__(self)
        self.meta_avail = False
        self.export_enabled = False
    
    @property
    def export_CASA_gaintable(self):
        return self.export_enabled
    
    @export_CASA_gaintable.setter
    def export_CASA_gaintable(self, value):
        self.export_enabled = value
    
    def set_metadata(self, src):
        """
            Sets up metadata taken from data provider
            
            Args:
                src: a cubical.data_handler instance
        """
        if six.PY3 and self.export_enabled:
            log.error("Gaintables cannot be written in Python 3 mode due to current casacore implementation issues")
            self.export_enabled = False

        if not isinstance(src, MSDataHandler):
            raise TypeError("src must be of type Cubical DataHandler")

        if self.export_enabled:
            self.antoffset = src._anttabcols["OFFSET"]
            self.antpos = src._anttabcols["POSITION"]
            self.anttype = src._anttabcols["TYPE"]
            self.antdishdiam = src._anttabcols["DISH_DIAMETER"]
            self.antflagrow = src._anttabcols["FLAG_ROW"]
            self.antmount = src._anttabcols["MOUNT"]
            self.antnames = src._anttabcols["NAME"]
            self.antstation = src._anttabcols["STATION"]
            self.fielddelaydirs = src._fldtabcols["DELAY_DIR"]
            self.fieldphasedirs = src._fldtabcols["PHASE_DIR"]
            self.fieldrefdir = src._fldtabcols["REFERENCE_DIR"]
            self.fieldcode = src._fldtabcols["CODE"]
            self.fieldflagrow = src._fldtabcols["FLAG_ROW"]
            self.fieldname = src._fldtabcols["NAME"]
            self.fieldnumpoly = src._fldtabcols["NUM_POLY"]
            self.fieldsrcid = src._fldtabcols["SOURCE_ID"]
            self.fieldtime = src._fldtabcols["TIME"]
            self.obstimerange = src._obstabcols["TIME_RANGE"]
            self.obslog = src._obstabcols["LOG"]
            self.obsschedule = src._obstabcols["SCHEDULE"]
            self.obsflagrow = src._obstabcols["FLAG_ROW"]
            self.obsobserver = src._obstabcols["OBSERVER"]
            self.obsproject = src._obstabcols["PROJECT"]
            self.obsreleasedate = src._obstabcols["RELEASE_DATE"]
            self.obsscheduletype = src._obstabcols["SCHEDULE_TYPE"]
            self.obstelescopename = src._obstabcols["TELESCOPE_NAME"]
            self.spwmeasfreq = src._spwtabcols["MEAS_FREQ_REF"]
            self.spwchanfreq = src._spwtabcols["CHAN_FREQ"]
            self.spwreffreq = src._spwtabcols["REF_FREQUENCY"]
            self.spwchanwidth = src._spwtabcols["CHAN_WIDTH"]
            self.spweffbw = src._spwtabcols["EFFECTIVE_BW"]
            self.spwresolution = src._spwtabcols["RESOLUTION"]
            self.spwflagrow = src._spwtabcols["FLAG_ROW"]
            self.spwfreqgroup = src._spwtabcols["FREQ_GROUP"]
            self.spwfreqgroupname = src._spwtabcols["FREQ_GROUP_NAME"]
            self.spwifconvchain = src._spwtabcols["IF_CONV_CHAIN"]
            self.spwname = src._spwtabcols["NAME"]
            self.spwnetsideband = src._spwtabcols["NET_SIDEBAND"]
            self.spwnumchan = src._spwtabcols["NUM_CHAN"]
            self.spwtotalbandwidth = src._spwtabcols["TOTAL_BANDWIDTH"]
            self.ddid_spw_map = src._ddid_spw
            self.sel_ddids = src._ddids
            self.do_write_casatbl = True
            self.meta_avail = True
        else:
            self.meta_avail = False
            

    def __export(self):
        """ exports the database to CASA gaintables """
        if six.PY3:
            log.error("Gaintables cannot be written in Python 3 mode due to current casacore implementation issues")
            return
        self._load(self.filename)
        
        if not self.meta_avail:
            log.error("Measurement metadata not set. Cannot write CASA-style gaintable.")
        elif not self.export_enabled:
            log.warning("By user request not exporting CASA-style gaintables.")
        else:
            log.info("Exporting to CASA gaintables")
            if "G:phase" in self.names():
                assert "G:phase.err" in self.names(), "Gain phase error not present in solutions db? This is a bug"
                casa_caltable_factory.create_G_table(self, "G:phase")
            if "G:gain" in self.names():
                assert "G:gain.err" in self.names(), "Gain error not present in solutions db? This is a bug"
                casa_caltable_factory.create_B_table(self, "G:gain")
                casa_caltable_factory.create_D_table(self, "G:gain")
            if "G:delay" in self.names():
                assert "G:delay.err" in self.names(), "Delay error not present in solutions db? This is a bug"
                casa_caltable_factory.create_K_table(self, "G:delay")
                
        self.close()
            
    def close(self):
        """ see iface_database.close() for details """
        # move to closed state before exporting and loading back and sorting data
        do_export = (self.mode is "create")
        PickledDatabase.close(self) 
        if do_export and self.export_enabled:
            self.__export()
    



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Creates summary plots using the solver stats.
"""
from __future__ import print_function
from builtins import range
import numpy as np
from cubical.tools import logger
log = logger.getLogger("plots")

from cubical.plots import DPI, ZOOM

def make_stats_plots(st, GD, basename, metadata):
    from cubical.plots import make_antenna_xaxis

    def save_figure(name, width, height):
        import pylab
        pylab.gcf().set_size_inches(min(width, 10000 / DPI), min(height, 10000 / DPI))
        filename = "{}.{}.png".format(basename, name)
        pylab.savefig(filename, dpi=DPI)
        print("saved plot " + filename, file=log)
        if GD["out"]["plots"] == "show":
            pylab.show()
        pylab.figure()

    import pylab

    # plot noise per time/channel
    pylab.subplot(121)
    pylab.title("Noise on input data")
    pylab.xlabel("channel")
    pylab.ylabel("timeslot")
    st.timechan.dv2[st.timechan.dv2 == 0] = np.inf
    pylab.imshow(np.sqrt(st.timechan.dv2), aspect='auto')
    pylab.colorbar()
    pylab.subplot(122)
    pylab.title("Noise on residuals")
    pylab.xlabel("channel")
    pylab.ylabel("timeslot")
    st.timechan.dr2[st.timechan.dr2 == 0] = np.inf
    pylab.imshow(np.sqrt(st.timechan.dr2), aspect='auto')
    pylab.colorbar()
    nt, nf = st.timechan.dv2.shape
    save_figure("noise.tf", max(nf * ZOOM / DPI * 2.5, 10), max(nt * ZOOM / DPI * 1.1,8))

    # plot chi-sq per time/channel
    pylab.subplot(121)
    pylab.title("Chi-sq on input data")
    pylab.xlabel("channel")
    pylab.ylabel("timeslot")
    st.timechan.initchi2[st.timechan.initchi2 == 0] = np.nan
    st.timechan.chi2[st.timechan.chi2 == 0] = np.nan
    pylab.imshow(st.timechan.initchi2, aspect='auto')
    pylab.colorbar()
    pylab.subplot(122)
    pylab.title("Chi-sq on residuals")
    pylab.xlabel("channel")
    pylab.ylabel("timeslot")
    pylab.imshow(st.timechan.chi2, aspect='auto')
    pylab.colorbar()
    nt, nf = st.timechan.chi2.shape
    save_figure("chi2.tf", max(nf * ZOOM / DPI * 2.5, 10), max(nt * ZOOM / DPI * 1.1, 8))

    # plot noise per antenna/channel
    pylab.subplot(121)
    noise = np.sqrt(st.chanant.dr2)
    noise[noise == 0] = np.inf
    nf, nant = noise.shape
    for ant in range(nant):
        pylab.plot(noise[:, ant], 'o-')
    for x in pylab.xticks()[0]:
        pylab.axvline(x, c="grey", lw=.5, ls=':', zorder=999)
    pylab.title("Noise (colour: antenna)")
    pylab.xlabel("channel")
    pylab.ylabel("noise")
    pylab.subplot(122)
    make_antenna_xaxis(metadata.antenna_name)
    for chan in range(nf):
        pylab.plot(noise[chan, :], 'o-')
    pylab.title("Noise (colour: channel)")
    pylab.ylabel("noise")
    save_figure("noise.antchan", 10, 5)

    # plot chi2 per antenna/channel
    pylab.subplot(121)
    chi2 = st.chanant.chi2
    chi2[chi2 == 0] = np.inf
    nf, nant = chi2.shape
    for ant in range(nant):
        pylab.plot(chi2[:, ant], 'o-')
    for x in pylab.xticks()[0]:
        pylab.axvline(x, c="grey", lw=.5, ls=':', zorder=999)
    pylab.title("Chi-sq (colour: antenna)")
    pylab.xlabel("channel")
    pylab.ylabel("$\chi^2$")
    pylab.subplot(122)
    make_antenna_xaxis(metadata.antenna_name)
    for chan in range(nf):
        pylab.plot(chi2[chan, :], 'o-')
    pylab.title("Chi-sq (colour: channel)")
    pylab.ylabel("$\chi^2$")
    save_figure("chi2.antchan", 10, 5)




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details

from __future__ import print_function
import six
try:
    from DDFacet.Imager import ClassDDEGridMachine
    if six.PY3:
        import DDFacet.cbuild.Gridder._pyGridderSmearPolsClassic3x as _pyGridderSmearPolsClassic
    else:
        import DDFacet.cbuild.Gridder._pyGridderSmearPolsClassic27 as _pyGridderSmearPolsClassic
    from DDFacet.ToolsDir.ModToolBox import EstimateNpix
    from DDFacet.Array import shared_dict
    from DDFacet.ToolsDir import ModFFTW
    import DDFacet.ToolsDir.ModTaper as MT
except ImportError:
    raise ImportError("Could not import DDFacet")

from . import FITSBeamInterpolator 
from . import DicoSourceProvider
from cubical.tools import logger, ModColor
log = logger.getLogger("DDFacetSim")
import numpy as np
import math
import os
from concurrent.futures import (ProcessPoolExecutor as PPE,
                                ThreadPoolExecutor as TPE)
import hashlib
import datetime as dt
import pyrap.quanta as pq
from numba import jit, prange
import concurrent.futures as cf
import cubical.workers as cw
import multiprocessing
from astropy import wcs, coordinates, units

def init_degridders(dir_CFs, 
                    dir_DCs,
                    CFs_dict, 
                    freqs, 
                    subregion_index, 
                    lmShift, 
                    dname, 
                    nfreqbands, 
                    DataCorrelationFormat,
                    sems,
                    npix,
                    should_init_cf,
                    wmax,
                    GD,
                    gmachines=None):
    """ Dummy initialization method to be executed on process pool """
    if cw.num_threads > 1:
        old_omp_setting = os.environ.get("OMP_NUM_THREADS", str(multiprocessing.cpu_count()))
        # OMP is not fork safe.... we have to force serialization in case the solvers are already set up
        # to use threading
        os.environ["OMP_NUM_THREADS"] = "1"
    dir_CFs = dir_CFs.instantiate()
    dir_DCs = dir_DCs.instantiate()
    CFs_dict = CFs_dict.instantiate()
    gmach = ClassDDEGridMachine.ClassDDEGridMachine(GD,
                ChanFreq = freqs,
                Npix = npix,
                lmShift = lmShift,
                IDFacet = dir_CFs[dname][subregion_index],
                SpheNorm = False, # Depricated, set ImToGrid True in .get!!
                NFreqBands = nfreqbands,
                DataCorrelationFormat=DataCorrelationFormat,
                ExpectedOutputStokes=[1], # Stokes I
                ListSemaphores=sems,
                cf_dict=CFs_dict,
                compute_cf=should_init_cf,
                wmax=wmax)
    (gmachines is not None) and gmachines.append(gmach)
    if should_init_cf:
        wnd_detaper = MT.Sphe2D(npix)
        wnd_detaper[wnd_detaper != 0] = 1.0 / wnd_detaper[wnd_detaper != 0]
        dir_DCs["facet_{}".format(dir_CFs[dname][subregion_index])] = wnd_detaper

    if cw.num_threads > 1:
        os.environ["OMP_NUM_THREADS"] = old_omp_setting

global UNIQ_ID
try:
    UNIQ_ID
except NameError:
    UNIQ_ID = os.getpid()

class DDFacetSim(object):
    __degridding_semaphores = None
    __initted_CF_directions = []
    __direction_CFs = shared_dict.SharedDict("facetids.cubidegrid.{}".format(UNIQ_ID))
    __ifacet = 0
    __CF_dict = {}
    __jones_cache = {}
    __should_init_sems = True
    __detaper_cache = shared_dict.SharedDict("tapers.cubidegrid.{}".format(UNIQ_ID))
    __exec_pool = None
    __IN_PARALLEL_INIT = False
    def __init__(self, num_processes=0):
        """ 
            Initializes a DDFacet model predictor
        """
        self.__direction = None
        self.__model = None
        self.init_sems()

    @classmethod
    def clean_jones_cache(cls):
        cls.__jones_cache = {}

    @classmethod
    def initialize_pool(cls, num_processes=0):
        if num_processes > 1:
            if DDFacetSim.__exec_pool is None:
                DDFacetSim.__exec_pool = PPE(max_workers=num_processes)
            DDFacetSim.__IN_PARALLEL_INIT = True

    @classmethod
    def shutdown_pool(cls):
        if DDFacetSim.__IN_PARALLEL_INIT:
            DDFacetSim.__exec_pool.shutdown()

    def set_direction(self, val):
        """ sets the direction in the cubical model cube to pack model data into """
        self.__direction = val
    
    def set_model_provider(self, model):
        """ sets the model provider """
        if not isinstance(model, DicoSourceProvider.DicoSourceProvider):
            raise TypeError("Model provider must be a DicoSourceProvider")
        log(2).print("Model predictor is switching to model '{0:s}'".format(str(model)))
        self.__model = model

    @classmethod
    def dealloc_degridders(cls):
        """ resets CF allocation state of all degridders """
        cls.__initted_CF_directions = []
        cls.__direction_CFs.delete() # cleans out /dev/shm allocated memory
        cls.__ifacet = 0
        for cf in cls.__CF_dict:
            cls.__CF_dict[cf].delete() # cleans out /dev/shm allocated memory
        cls.__CF_dict = {}
        cls.__should_init_sems = True
        cls.__detaper_cache.delete() # cleans out /dev/shm allocated memory

    @classmethod
    def init_sems(cls, NSemaphores = 3373):
        """ Init semaphores """
        if not cls.__should_init_sems:
            return
        cls.__should_init_sems = False
        cls.__degridding_semaphores = ["Semaphore.cubidegrid{0:d}".format(i) for i in
                                        range(NSemaphores)]
        _pyGridderSmearPolsClassic.pySetSemaphores(cls.__degridding_semaphores)
    
    @classmethod
    def del_sems(cls):
        """ Deinit semaphores """
        if not cls.__should_init_sems:
            return
        cls.__should_init_sems = True
        _pyGridderSmearPolsClassic.pyDeleteSemaphore()
        cls.__degridding_semaphores = None

    def bandmapping(self, vis_freqs, nbands):
        """ 
            Gives the frequency mapping for visibility to degrid band 
            For now we assume linear regular spacing, so we may end up
            with a completely empty band somewhere in the middle
            if we have spws that are separated in frequency
        """
        band_frequencies = np.linspace(np.min(vis_freqs), np.max(vis_freqs), nbands)
        def find_nearest(array, value):
            idx = np.searchsorted(array, value, side="left")
            if idx > 0 and (idx == len(array) or math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
                return idx - 1
            else:
                return idx
        freq_mapping = [find_nearest(band_frequencies, v) for v in vis_freqs]
        return band_frequencies, freq_mapping

    def __mjd2dt(self, utc_timestamp):
        """
        Converts array of UTC timestamps to list of datetime objects for human readable printing
        """
        return [dt.datetime.utcfromtimestamp(pq.quantity(t, "s").to_unix_time()) for t in utc_timestamp]

    def radec2lm_scalar(self,ra,dec):
        """ 
            SIN projected world to l,m projection 
            AIPS memo series, 27
            Eric Greisen
        """
        ra0, dec0 = self.__phasecenter
        l = np.cos(dec) * np.sin(ra - ra0)
        m = np.sin(dec) * np.cos(dec0) - np.cos(dec) * np.sin(dec0) * np.cos(ra - ra0)
        return l,m

    def __apply_jones(self, jones_matrix, vis, a0, a1):
        """
            Elementwise apply J1 . V12 . J2^H to degridded model data
            Parameters: 
                jones_matrix: Jones Matrix containing one direction to apply for this facet
                              as returned by __load_jones
                vis: row x nchan x ncorr matrix with ncorr == 4
                a0: antenna1 index (not name) per row, as defined in CASA memo 229
                a1: antenna1 index (not name) per row, as defined in CASA memo 229

                jones_matrix: must have a time and frequency mapping to the space jones matricies
                              included. This should be computed by __load_jones
        """
        @jit(nopython=True, nogil=True, cache=True, parallel=True)
        def __JpXJqH(vis, out_vis, jones, tmap, numap, a0, a1, idir=0):
            """ element wise onion multiply for visibilities of shape (row x chan x corr) """
            nrow, nch, ncorr = vis.shape
            for r in prange(nrow):
                jt_indx = tmap[r]
                j_a1indx = a0[r] 
                j_a2indx = a1[r]
                for ch in range(nch):
                    jnu_indx = numap[ch]
                    # ntime, ndir, na, nnu, 2x2
                    jp = jones[jt_indx, idir, j_a1indx, jnu_indx, :, :]
                    jq = jones[jt_indx, idir, j_a2indx, jnu_indx, :, :]
                    X = vis[r, ch, :]
                    # jq^H
                    a = jq[0, 0].conjugate()
                    b = jq[1, 0].conjugate()
                    c = jq[0, 1].conjugate()
                    d = jq[1, 1].conjugate()
                    # X
                    XX = X[0]
                    XY = X[1]
                    YX = X[2]
                    YY = X[3]
                    # jp
                    e = jp[0, 0]
                    f = jp[0, 1]
                    g = jp[1, 0]
                    h = jp[1, 1]
                    # jp.X.jq^H
                    out_vis[r, ch, 0] = e * (XX*a + XY*c) + f * (YX*a + YY*c)
                    out_vis[r, ch, 1] = e * (XX*b + XY*d) + f * (YX*b + YY*d)
                    out_vis[r, ch, 2] = g * (XX*a + XY*c) + h * (YX*a + YY*c)
                    out_vis[r, ch, 3] = g * (XX*b + XY*d) + h * (YX*b + YY*d)
        
        if jones_matrix is None:
            return vis.copy() # unity case
        tmap = jones_matrix["DicoJones_Beam"]["TimeMapping"]
        numap = jones_matrix["DicoJones_Beam"]["Jones"]["VisToJonesChanMapping"]
        jones = jones_matrix["DicoJones_Beam"]["Jones"]["Jones"]
        assert jones.ndim == 6
        jnt, jndir, jnA, jnnu, jnc1, jnc2 = jones.shape
        nvisrow, nvischan, nviscorr = vis.shape
        assert nviscorr == 4
        assert jndir == 1 # one jones stack per facet
        assert tmap.size == vis.shape[0] # time map per row
        assert numap.size == vis.shape[1] # chan map per channel
        assert a0.size == vis.shape[0]
        assert a1.size == vis.shape[0]

        apparent_scale_vis = np.zeros_like(vis, dtype=vis.dtype)
        __JpXJqH(vis, apparent_scale_vis, jones, tmap, numap, a0, a1) 
        return apparent_scale_vis

    def __load_jones(self,
                     src, 
                     sub_region_index,
                     field_centre, 
                     chunk_ctr_frequencies, 
                     chunk_channel_widths,
                     correlation_ids,
                     station_positions,
                     utc_times,
                     station_names,
                     provider=None):
        """
            src: current dico source provider
            sub_region_index: source provider facet index
            field_centre: radian ra dec coordinate of the original field phase centre
            chunk_ctr_frequencies: center channel frequencies (Hz) of chunk to predict E terms for
            chunk_channel_widths: channel widths of chunk channels
            correlation_ids: hand labels per correlation term as defined in casacore Stokes.h
            station_positions: ECEF coordinates for stations (na, 3) array
            utc_times: UTC times as specified by measurement set
            station_names: station names as loaded from measurement set
            provider: FITS or None available at present
        """
        src_name = self.__cachename_compute(src)
        def __beam_cache_name(field_centre, chunk_ctr_frequencies, correlation_ids, station_positions, utc_times):
            res = ":".join([",".join(map(str, field_centre)),
                            ",".join(map(str, [np.min(chunk_ctr_frequencies), np.max(chunk_ctr_frequencies)])),
                            str(chunk_ctr_frequencies.size),
                            ",".join(map(str, correlation_ids)),
                            ",".join(map(str, station_positions)),
                            ",".join(map(str, [np.min(utc_times), np.max(utc_times)])),
                            str(src.direction),
                            str(sub_region_index)
                           ])
            return hashlib.sha512(res.encode()).hexdigest()
        
        def __vis_2_chan_map(freq, FreqDomains):
            """Builds mapping from tile frequency axis to 
               frequency axis (FreqDomains) as defined by the FITS interpolator
            """
            NChanJones = FreqDomains.shape[0]
            MeanFreqJonesChan = (FreqDomains[:, 0]+FreqDomains[:, 1])/2.
            DFreq = np.abs(freq.reshape(
                (freq.size, 1))-MeanFreqJonesChan.reshape((1, NChanJones)))
            return np.argmin(DFreq, axis=1)

        def __vis_2_time_map(DicoSols, times):
            """Builds mapping from MS rows to Jones solutions.
            Args:
                DicoSols: dictionary of Jones matrices, which includes t0 and t1 entries

            Returns:
                Vector of indices, one per each row in DATA, giving the time index of the Jones matrix
                corresponding to that row.
            """
            DicoJonesMatrices = DicoSols
            ind = np.zeros((len(times),), np.int32)
            ii = 0
            assert DicoJonesMatrices["t0"].size == DicoJonesMatrices["t1"].size # same number of jones bin start and end times
            for it in range(DicoJonesMatrices["t0"].size):
                t0 = DicoJonesMatrices["t0"][it]
                t1 = DicoJonesMatrices["t1"][it]
                ind[(times >= t0) & (times < t1)] = it
            return ind

        beam_cache_name = __beam_cache_name(field_centre, chunk_ctr_frequencies, 
                                            correlation_ids, station_positions, utc_times)
        if beam_cache_name not in self.__jones_cache:
            dt_start = self.__mjd2dt([np.min(utc_times)])[0].strftime('%Y/%m/%d %H:%M:%S')
            dt_end = self.__mjd2dt([np.max(utc_times)])[0].strftime('%Y/%m/%d %H:%M:%S')
            if provider is None:
                log(2).print("Setting E Jones to unity for '{0:s}' direction '{1:s}' facet '{2:d}' between {3:s} and {4:s} UTC for "
                             "fractional bandwidth {5:.2f}~{6:.2f} MHz".format(
                             str(self.__model), str(self.__direction), sub_region_index, dt_start, dt_end, 
                             np.min(chunk_ctr_frequencies*1e-6), np.max(chunk_ctr_frequencies*1e-6)))
                jones_matrix = None
            elif provider == "FITS":
                log(2).print("Calculating E Jones for '{0:s}' direction '{1:s}' facet '{2:d}' between {3:s} and {4:s} UTC for "
                             "fractional bandwidth {5:.2f}~{6:.2f} MHz".format(
                             str(self.__model), str(self.__direction), sub_region_index, dt_start, dt_end, 
                             np.min(chunk_ctr_frequencies*1e-6), np.max(chunk_ctr_frequencies*1e-6)))
                beam_provider = FITSBeamInterpolator.FITSBeamInterpolator(field_centre,
                                                                          chunk_ctr_frequencies,
                                                                          chunk_channel_widths,  
                                                                          correlation_ids,
                                                                          station_positions,
                                                                          station_names,
                                                                          self.degrid_opts)
                jones_matrix = {"DicoJones_Beam": {
                        "Dirs": {},
                        "TimeMapping": {},
                        "Jones": {},
                        "TimeMapping": {},
                        "MapJones": {},
                        "FITSProvider": beam_provider
                    }
                }
                subregion_pixoffset = src.get_direction_pxoffset(subregion_index=sub_region_index)
                header = {
                    "NAXIS": 2,
                    "NAXIS1": subregion_pixoffset[0],
                    "CRVAL1": np.rad2deg(self.__phasecenter[0]),
                    "CRPIX1": 0,
                    "CDELT1": src.pixel_scale/3600.0, # already negated in offset calc
                    "CUNIT1": "deg",
                    "CTYPE1": "RA---SIN",
                    "NAXIS2": subregion_pixoffset[1],
                    "CRVAL2": np.rad2deg(self.__phasecenter[1]),
                    "CRPIX2": 0,
                    "CDELT2": src.pixel_scale/3600.0,
                    "CUNIT2": "deg",
                    "CTYPE2": "DEC--SIN",
                }
                thiswcs = wcs.WCS(header)
                sky = thiswcs.pixel_to_world(subregion_pixoffset[0], subregion_pixoffset[1])
                ra, dec = map(np.deg2rad, [sky.ra.value, sky.dec.value])
                l, m = self.radec2lm_scalar(ra, dec)
                
                jones_matrix["DicoJones_Beam"]["Dirs"]["l"] = l
                jones_matrix["DicoJones_Beam"]["Dirs"]["m"] = m
                jones_matrix["DicoJones_Beam"]["Dirs"]["ra"] = ra # radians
                jones_matrix["DicoJones_Beam"]["Dirs"]["dec"] = dec # radians
                I = np.sum(np.mean(src.get_degrid_model(subregion_index=sub_region_index)[:,0,:,:], axis=0))
                jones_matrix["DicoJones_Beam"]["Dirs"]["I"] = I
                jones_matrix["DicoJones_Beam"]["Dirs"]["Cluster"] = 0 # separate matrix stack for each subregion (facet)
                sample_times = np.array(beam_provider.getBeamSampleTimes(utc_times))
                dt = np.zeros_like(sample_times)
                if sample_times.size > 1:
                    # half width of beam sample time
                    dt[:-1] = (sample_times[1:] - sample_times[:-1]) / 2.
                    dt[-1] = dt[-2] # equal-sized left and right dt of the last step
                else:
                    dt[...] = 1.0e-6
                t0 = sample_times - dt
                t1 = sample_times + dt
                E_jones = np.ascontiguousarray([beam_provider.evaluateBeam(t, [ra], [dec]) for t in sample_times], dtype=np.complex64)
                jones_matrix["DicoJones_Beam"]["Jones"]["t0"] = np.ascontiguousarray(t0, dtype=np.float64) # lower time boundary of beam solution
                jones_matrix["DicoJones_Beam"]["Jones"]["t1"] = np.ascontiguousarray(t1, dtype=np.float64) # upper time boundary of beam solution
                jones_matrix["DicoJones_Beam"]["Jones"]["tm"] = np.ascontiguousarray((t0 + dt), dtype=np.float64) # beam time mid point
                jones_matrix["DicoJones_Beam"]["Jones"]["Jones"] = E_jones
                jones_matrix["DicoJones_Beam"]["Jones"]["VisToJonesChanMapping"] = __vis_2_chan_map(chunk_ctr_frequencies, beam_provider.getFreqDomains())
                jones_matrix["DicoJones_Beam"]["TimeMapping"] = __vis_2_time_map(jones_matrix["DicoJones_Beam"]["Jones"], 
                                                                                 utc_times)
                jones_matrix["DicoJones_Beam"]["SampleFreqs"] = beam_provider.getFreqDomains()
                #import ipdb; ipdb.set_trace()
            else:
                raise ValueError("E Jones provider '{0:s}' not understood. Only FITS or None available".format(provider))

            DDFacetSim.__jones_cache[beam_cache_name] = jones_matrix
        return DDFacetSim.__jones_cache[beam_cache_name]

    def __cachename_compute(self, src):
        reg_props = str(src.subregion_count)
        for subregion_index in range(src.subregion_count):
            reg_props += " npx:" + str(src.get_direction_npix(subregion_index=subregion_index)) + \
                    " scale:" + "x".join(map(str, list(np.deg2rad(src.get_direction_pxoffset(subregion_index=subregion_index)) *
                                                       src.pixel_scale / 3600.0)))
        res = "dir_{0:s}_{1:s}_{2:s}".format(str(self.__model), str(self.__direction), str(reg_props))
        return hashlib.sha512(res.encode()).hexdigest()

    def __init_grid_machine(self, src, dh, tile, poltype, freqs, chan_widths):
        """ initializes a grid machine for this direction """
        if self.__degridding_semaphores is None:
            raise RuntimeError("Need to initialize degridding semaphores first. Call init_sems")
        if poltype == "linear":
            DataCorrelationFormat = [9, 10, 11, 12] # Defined in casacore Stokes.h
        elif poltype == "circular":
            DataCorrelationFormat = [5, 6, 7, 8]
        else:
            raise ValueError("Only supports linear or circular for now")

        should_init_cf = self.__cachename_compute(src) not in self.__initted_CF_directions

        # Kludge up a DDF GD
        GD = dict(RIME={}, Facets={}, CF={}, Image={}, DDESolutions={}, Comp={}, Beam={})
        GD["RIME"]["Precision"] = "S"
        GD["Facets"]["Padding"] = dh.degrid_opts["Padding"]
        GD["CF"]["OverS"] = dh.degrid_opts["OverS"]
        GD["CF"]["Support"] = dh.degrid_opts["Support"]
        GD["CF"]["Nw"] = dh.degrid_opts["Nw"]
        GD["Image"]["Cell"] = src.pixel_scale
        GD["DDESolutions"]["JonesMode"] = "Full" 	  # #options:Scalar|Diag|Full
        GD["DDESolutions"]["Type"] = "Nearest" # Deprecated? #options:Krigging|Nearest
        GD["DDESolutions"]["Scale"] = 1.      # Deprecated? #metavar:DEG
        GD["DDESolutions"]["gamma"] = 4.	  # Deprecated? 
        GD["RIME"]["FFTMachine"] = "FFTW" 
        GD["Comp"]["BDAJones"] = 0         # If disabled, gridders and degridders will apply a Jones terms per visibility. 
                                           # If 'grid', gridder will apply them per BDA block, if 'both' so will the degridder. This is faster but possibly less 
                                           # accurate, if you have rapidly evolving Jones terms.
        GD["DDESolutions"]["DDModeGrid"] = "AP"	  # In the gridding step, apply Jones matrices Amplitude (A) or Phase (P) or Amplitude&Phase (AP)
        GD["DDESolutions"]["DDModeDeGrid"] = "AP"	  # In the degridding step, apply Jones matrices Amplitude (A) or Phase (P) or Amplitude&Phase (AP)
        GD["DDESolutions"]["ScaleAmpGrid"] = 0 # Deprecated?
        GD["DDESolutions"]["ScaleAmpDeGrid"] = 0	  # Deprecated?
        GD["DDESolutions"]["CalibErr"] = 10.	  # Deprecated?
        GD["DDESolutions"]["ReWeightSNR"] = 0.	  # Deprecated? 
        GD["RIME"]["BackwardMode"] = "BDA"
        GD["RIME"]["ForwardMode"] = "Classic" # Only classic for now... why would you smear unnecessarily in a model predict??
        self.degrid_opts = dh.degrid_opts
        # INIT degridder machine from DDF
        band_frequencies, freq_mapping = self.bandmapping(freqs, dh.degrid_opts["NDegridBand"])
        src.set_frequency(band_frequencies)
        wmax = np.max([dh.metadata.baseline_length[k] for k in dh.metadata.baseline_length])
        gmachines = []
        dname = self.__cachename_compute(src)
        if should_init_cf:
            log(2).print("This is the first time predicting for '{0:s}' direction '{1:s}'. "
                         "Initializing degridder for {2:d} facets - this may take a wee bit of time.".format(
                         str(self.__model), str(self.__direction), src.subregion_count))
            DDFacetSim.__initted_CF_directions.append(dname)    
            DDFacetSim.__direction_CFs[dname] = DDFacetSim.__direction_CFs.get(dname, []) + list(DDFacetSim.__ifacet + np.arange(src.subregion_count)) #unique facet index for this subregion
            DDFacetSim.__ifacet += src.subregion_count
            for sri in range(src.subregion_count):
                DDFacetSim.__CF_dict["{}.{}".format(dname, sri)] = \
                    shared_dict.SharedDict("convfilters.cubidegrid.{}.{}.{}".format(UNIQ_ID, dname, sri))

        # init in parallel
        futures = {}        
        for subregion_index in range(src.subregion_count):
            ra, dec = np.deg2rad(src.get_direction_pxoffset(subregion_index=subregion_index) * 
                                    src.pixel_scale / 3600.0) + self.__phasecenter
            l, m = self.radec2lm_scalar(ra, dec)
            init_args = (DDFacetSim.__direction_CFs.readwrite(), 
                        DDFacetSim.__detaper_cache.readwrite(),
                        DDFacetSim.__CF_dict["{}.{}".format(dname, subregion_index)].readwrite(),
                        freqs,
                        subregion_index,
                        np.array([l, m]),
                        dname,
                        dh.degrid_opts["NDegridBand"],
                        DataCorrelationFormat,
                        self.__degridding_semaphores,
                        src.get_direction_npix(subregion_index=subregion_index),
                        should_init_cf,
                        wmax,
                        GD)
            if DDFacetSim.__IN_PARALLEL_INIT:
                futures[DDFacetSim.__exec_pool.submit(init_degridders, *init_args)] = subregion_index
            else:
                init_degridders(*init_args)

        if DDFacetSim.__IN_PARALLEL_INIT:
            def reap_children():
                pid, status, _ = os.wait3(os.WNOHANG)
                if pid:
                    print("child process {} exited with status {}. This is a bug, or an out-of-memory condition.".format(pid, status), file=log(0,"red"))
                    print("This error is not recoverable: the main process will now commit ritual harakiri.", file=log(0,"red"))
                    os._exit(1)
            while futures:
                reap_children()
                done, not_done = cf.wait(list(futures.keys()), timeout=1)
                for future in done:
                    key = futures[future]
                    stats = future.result()
                    del futures[future]
        
        # construct handles after the initialization has been completed
        for subregion_index in range(src.subregion_count):
            init_degridders(DDFacetSim.__direction_CFs.readwrite(), 
                            DDFacetSim.__detaper_cache.readwrite(),
                            DDFacetSim.__CF_dict["{}.{}".format(dname, subregion_index)].readwrite(),
                            freqs,
                            subregion_index,
                            np.deg2rad(src.get_direction_pxoffset(subregion_index=subregion_index) * 
                                       src.pixel_scale / 3600.0),
                            dname,
                            dh.degrid_opts["NDegridBand"],
                            DataCorrelationFormat,
                            self.__degridding_semaphores,
                            src.get_direction_npix(subregion_index=subregion_index),
                            False, #already initialized
                            wmax,
                            GD,
                            gmachines
                            )
        return gmachines

    @classmethod
    def __detaper_model(cls, gm, model_image, idfacet):
        """ Detapers model image by the fourier inverse of the convolution kernel 
            Assertions that the tapering function is larger than the model facet,
            both are square and odd sized
        """
        assert model_image.shape[2] == model_image.shape[3]
        wnd_detaper = cls.__detaper_cache.readonly().instantiate()["facet_{}".format(idfacet)]
        assert wnd_detaper.shape[0] == model_image.shape[2] and wnd_detaper.shape[1] == model_image.shape[3]
        model_image.real *= wnd_detaper[None, None, :, :]
        return model_image

    def simulate(self, dh, tile, tile_subset, poltype, uvwco, freqs, model_type, chan_widths, utc_times):
        """ Predicts model data for the set direction of the dico source provider 
            returns a ndarray model of shape nrow x nchan x 4
        """
        if self.__direction is None:
            raise RuntimeError("Direction has not been set. Please set direction before simulating")
        if self.__model is None:
            raise RuntimeError("Model has not been set. Please set model before simulating")
        DDFacetSim.initialize_pool(dh.degrid_opts["NProcess"])
        if not isinstance(dh.degrid_opts["PointingCenterAt"], str) and not \
           (isinstance(dh.degrid_opts["PointingCenterAt"], list) and len(dh.degrid_opts["PointingCenterAt"]) == 3):
            raise RuntimeError("--degridding-PointingCenterAt must be a string or coordinate list. "
                               "It can be either: 'DataPhaseDir' or a 'j2000,rahms,decsdms'-style coordinate that specifies "
                               "the pointing centre of the model if it has been rephased (for e.g. mosaicing) prior to imaging")
        if isinstance(dh.degrid_opts["PointingCenterAt"], str) and \
           dh.degrid_opts["PointingCenterAt"].strip().upper() == "DATAPHASEDIR":
            self.__phasecenter = dh.phadir
        else:
            try:
                epoch,ra,dec = map(lambda x: x.strip(), dh.degrid_opts["PointingCenterAt"])
                if epoch.upper() != "J2000":
                    raise RuntimeError("Only supports j2000 coordinates for rephasing at present")
                crd = coordinates.SkyCoord(f"{ra} {dec}", frame="fk5", equinox=epoch)
                self.__phasecenter = np.array(list(map(np.deg2rad, [crd.ra.value, crd.dec.value])))
            except Exception as e:
                raise RuntimeError(f"Invalid dataset rephaseing coordinate specification. Expected j2000,ra,dec. "
                                   f"Detailed message: {str(e)}")
        freqs = freqs.ravel()
        src = self.__model
        src.set_direction(self.__direction)
        band_frequencies, freq_mapping = self.bandmapping(freqs, dh.degrid_opts["NDegridBand"])
        gmacs = self.__init_grid_machine(src, dh, tile, poltype, freqs, chan_widths.ravel())
        nrow = uvwco.shape[0]
        nfreq = len(freqs)
        if model_type not in ["cplx2x2", "cplxdiag", "cplxscalar"]:
            raise ValueError("Only supports cplx2x2 or cplxdiag or cplxscalar models at the moment")
        ncorr = 4
        region_model = np.zeros((nrow, nfreq, 4), dtype=np.complex64)
        flagged = np.zeros_like(region_model, dtype=np.bool)
        # now we predict for this direction
        log(2).print("\tComputing visibilities in {1:d} facets for direction '{0:s}' for model '{2:s}'...".format(
                     str(self.__direction), src.subregion_count, str(self.__model)))

        for gm, subregion_index in zip(gmacs, range(src.subregion_count)):
            model = np.zeros((nrow, nfreq, 4), dtype=np.complex64).copy()
            model_image = src.get_degrid_model(subregion_index=subregion_index).astype(dtype=np.complex64).copy() #degridder needs transposes, dont mod the data globally

            if not np.any(model_image):
                log(2).print("\tFacet {0:d} is empty. Skipping".format(subregion_index))
                continue
            dname = self.__cachename_compute(src)
            model_image = DDFacetSim.__detaper_model(gm, model_image.view(), self.__direction_CFs[dname][subregion_index]).copy() # degridder don't respect strides must be contiguous

            # apply normalization factors for FFT
            model_image[...] *= (model_image.shape[3] ** 2) * (gm.WTerm.OverS ** 2) 

            # load cached E Jones (None if unity is to be applied):
            if poltype == "linear":
                DataCorrelationFormat = list(map(lambda x: FITSBeamInterpolator.MS_STOKES_ENUMS.index(x), 
                                                 FITSBeamInterpolator.LINEAR_CORRS))
            elif poltype == "circular":
                DataCorrelationFormat = list(map(lambda x: FITSBeamInterpolator.MS_STOKES_ENUMS.index(x), 
                                                 FITSBeamInterpolator.CIRCULAR_CORRS))
            else:
                raise ValueError("Only supports linear or circular feeds for now")

            E_jones = self.__load_jones(src = src,
                                        sub_region_index = subregion_index,
                                        field_centre = self.__phasecenter,
                                        chunk_ctr_frequencies = freqs.ravel(),
                                        chunk_channel_widths = chan_widths.ravel(),
                                        correlation_ids = DataCorrelationFormat,
                                        station_positions = dh.antpos,
                                        station_names = dh.antnames,
                                        utc_times = utc_times,
                                        provider=dh.degrid_opts["BeamModel"])
            if cw.num_threads > 1:
                old_omp_setting = os.environ.get("OMP_NUM_THREADS", str(multiprocessing.cpu_count()))
                # OMP is not fork safe.... we have to force serialization in case the solvers are already set up
                # to use threading
                os.environ["OMP_NUM_THREADS"] = "1"
            
            this_region_model = -1 * gm.get( #default of the degridder is to subtract from the previous model
                times=tile_subset.time_col, 
                uvw=uvwco.astype(dtype=np.float64).copy(), 
                visIn=model, 
                flag=flagged, 
                A0A1=[tile_subset.antea.astype(dtype=np.int32).copy(), tile_subset.anteb.astype(dtype=np.int32).copy()], 
                ModelImage=model_image, 
                PointingID=src.direction,
                Row0Row1=(0, -1),
                DicoJonesMatrices=None, 
                freqs=freqs.astype(dtype=np.float64).copy(), 
                ImToGrid=False,
                TranformModelInput="FT", 
                ChanMapping=np.array(freq_mapping, dtype=np.int32), 
                sparsification=None)

            region_model += self.__apply_jones(vis = this_region_model,
                                               jones_matrix = E_jones,
                                               a0 = tile_subset.antea,
                                               a1 = tile_subset.anteb)
            
            if cw.num_threads > 1:
                os.environ["OMP_NUM_THREADS"] = old_omp_setting

        model_corr_slice = np.s_[0] if model_type == "cplxscalar" else \
                           np.s_[0::3] if model_type == "cplxdiag" else \
                           np.s_[:] # if model_type == "cplx2x2"
        if not np.any(region_model[:, :, model_corr_slice]):
            log.critical("Model in region '{0:s}' is completely empty. This may indicate user "
                         "error and lead to non-correcting direction dependent gains!".format(self.__direction))

        return region_model[:, :, model_corr_slice]

import atexit

def cleanup_degridders():
    DDFacetSim.del_sems()
    DDFacetSim.dealloc_degridders()
    DDFacetSim.clean_jones_cache()
    DDFacetSim.shutdown_pool()
        
atexit.register(cleanup_degridders)



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Cython kernels for various rebinning operations. Common dimensions of arrays are:

+----------------+------+
| Dimension      | Size |
+================+======+
| Direction      |   d  |
+----------------+------+
| Model          |   m  |
+----------------+------+
| Time           |   t  |
+----------------+------+
| Time Intervals |   ti |
+----------------+------+
| Frequency      |   f  |
+----------------+------+
| Freq Intervals |   fi |
+----------------+------+
| Antenna        |   a  |
+----------------+------+
| Correlation    |   c  |
+----------------+------+

"""
from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def rebin_index_columns(time, time0, antea, antea0, anteb, anteb0,
             			ddid_col, ddid_col0, row_map):
    """
    Rebins indexing columns (time, ant1/2, ddid), using the given row map.
    Time should be a zero-filled output column of the expected length (row_map[-1]+1)
    anteab, anteb, ddid_col should be empty output columns of the expected length.
    """

    n_input_rows = time0.shape[0]
    n_output_rows  = time.shape[0]

    vis_count = np.zeros(n_output_rows, np.int64)

    for input_row in range(n_input_rows):
        output_row = abs(row_map[input_row])
        vis_count[output_row] += 1
        time[output_row] += time0[input_row]
        antea[output_row] = antea0[input_row]
        anteb[output_row] = anteb0[input_row]
        ddid_col[output_row] = ddid_col0[input_row]

    for output_row in range(n_output_rows):
        if vis_count[output_row]:
            time[output_row] /= vis_count[output_row]

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def rebin_vis(vis, vis0, uvw, uvw0, flag, flag0, weights, weights0, num_weights,
              rebin_row_map, rebin_chan_map):
    """
    Rebin the input data.
    """
    n_input_rows, n_fre0, n_cor0 = vis0.shape[0], vis0.shape[1], vis0.shape[2]
    n_row, n_fre, n_cor = vis.shape[0], vis.shape[1], n_cor0

    sum_ww = np.zeros((n_row,n_fre,n_cor), np.float32)    # Sum of weights per each output visibility
    sum_rw = np.zeros(n_row, np.float32)                  # Sum of weights per each output row

    for row0 in range(n_input_rows):
        row = rebin_row_map[row0]
        conjugate = row<0
        if conjugate:
            row = -row

        row_sum_weights = 0  # sum of weights of current _input_ row

        for f0 in range(n_fre0):
            f = rebin_chan_map[f0]
            for c in range(n_cor0):
		# Output flags accumulate all input flags across the bin with bitwise-OR.
		# However, below we'll clear them if at least one unflagged visibility was present.
                flag[row, f, c] |= flag0[row0, f0, c]  
                if not flag0[row0, f0, c]:
                    # accumulate weights
                    if num_weights:
                        for w in range(num_weights):
                            weights[w, row, f, c] += weights0[w, row0, f0, c]
                        ww = weights0[0, row0, f0, c]
                    else:
                        ww = 1
                    vis[row, f, c] += ww*vis0[row0, f0, c].conjugate() if conjugate else ww*vis0[row0, f0, c]
                    sum_ww[row, f, c] += ww
                    sum_rw[row] += ww
                    row_sum_weights += ww
        if row_sum_weights:
            for i in range(3):
                uvw[row, i] += row_sum_weights*uvw0[row0, i]

    # Now normalize by counts and sums of the weights.
    for row in range(n_row):
        if sum_rw[row]:
            for i in range(3):
                uvw[row, i] /= sum_rw[row]
            for f in range(n_fre):
                for c in range(n_cor0):
                    if sum_ww[row, f, c]:
                        flag[row, f, c] = 0
                        vis[row, f, c] /= sum_ww[row, f, c]

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def rebin_model(model, model0, flag0, rebin_row_map, rebin_chan_map):
    """
    Rebins a model column, following a rebin_row_map precomputed by cyrebin_vis
    """
    n_row0, n_fre0, n_cor0 = model0.shape[0], model0.shape[1], model0.shape[2]
    n_row, n_fre, n_cor = model.shape[0], model0.shape[1], n_cor0

    counts = np.zeros((n_row,n_fre,n_cor), np.int64)    # counts per each model visibility

    for row0 in range(n_row0):
        row = rebin_row_map[row0]
        conjugate = row<0
        if conjugate:
            row = -row

        for f0 in range(n_fre0):
            f = rebin_chan_map[f0]
            for c in range(n_cor0):
                if not flag0[row0, f0, c]:
                    counts[row, f, c] += 1
                    model[row, f, c] += model0[row0, f0, c].conjugate() if conjugate else model0[row0, f0, c]

    # now normalize by counts
    for row in range(n_row):
        for f in range(n_fre):
            for c in range(n_cor):
                if counts[row, f, c]:
                    model[row, f, c] /= counts[row, f, c]





from __future__ import print_function
from builtins import range
import math,cmath
import numpy as np
import numpy.ma as ma
from cubical.tools import logger
log = logger.getLogger("plots")
from past.builtins import cmp

from cubical.plots import DPI, ZOOM, make_antenna_xaxis

def _abs(x):
    """
    Works around numpy bug with abs() of masked arrays producing a 
    ComplexWarning(Casting complex values to real discards the imaginary part)
    """
    if ma.isMA(x):
        return ma.masked_array(np.abs(x.data), x.mask)
    else:
        return ma.masked_array(np.abs(x))
    


def _cmp_antenna (sa, sb):
    """Helper function to sort antenna names. Try numeric compare first, fall back to text compare if failed""";
    try:
        return cmp(int(sa), int(sb));
    except:
        return cmp(sa, sb);


def _normifrgain(rr):
    """Converts gains to mean offsets with std"""
    if type(rr) in (float, complex):
        return abs(rr), 0
    else:
        ## some np versions produce a ComplexWarning here because the fill_value stays complex
        # offset = np.abs(rr[rr != 1])
        offset = _abs(rr)
        offset[rr==1] = ma.masked
        if offset.count():
            with np.errstate(under='ignore'): # work around https://github.com/numpy/numpy/issues/4895
                return float(offset.mean()), float(offset.std())
        else:
            return 1, 0


def _complexifrgain(rr):
    """Converts gains to complex offsets with std"""
    if type(rr) in (float, complex):
        return rr, 0
    else:
        vals = rr[rr != 1]
        if vals.count():
            offset = float(_abs(vals).mean())
            mean = vals.mean().ravel()[0]
            mean = cmath.rect(offset, cmath.phase(mean))
            with np.errstate(under='ignore'): # work around https://github.com/numpy/numpy/issues/4895
                return mean, _abs(vals - mean).std()
        else:
            return 1,0


def _is_unity(rr, ll):
    return (not rr.count() or (rr==1).all()) and \
           (not ll.count() or (ll==1).all())


#def make_ifrgain_plots(filename="$STEFCAL_DIFFGAIN_SAVE", prefix="IG", feed="$IFRGAIN_PLOT_FEED", msname="$MS"):
def make_ifrgain_plots(ig, ms, GD, basename):
    """Makes a set of ifrgain plots from the specified saved file."""
    print("generating plots for suggested baseline-based corrections (BBCs)", file=log(0))

    metadata = ms.metadata
    import pylab
    def save_figure(name, width, height):
        import pylab
        pylab.gcf().set_size_inches(min(width, 10000 / DPI), min(height, 10000 / DPI))
        filename = "{}.{}.png".format(basename, name)
        pylab.savefig(filename, dpi=DPI)
        print("saved plot " + filename, file=log)
        if GD["out"]["plots"] == "show":
            pylab.show()
        pylab.figure()

    # load baseline info, if MS is available
    antpos = list(zip(ms.antnames, ms.antpos))
    # make dictionary of IFR name: baseline length
    baseline = { metadata.baseline_name[p,q]: metadata.baseline_length[p,q] 
                 for p in range(ms.nants) for q in range(p+1, ms.nants) }
    feeds = [ (x+y).upper() for x in metadata.feeds for y in metadata.feeds ]

    def plot_xy(content, title):
        """Plots x vs y"""
        pylab.errorbar(
            [x for l, (x, xe), (y, ye) in content], [y for l, (x, xe), (y, ye) in content],
            [ye for l, (x, xe), (y, ye) in content], [xe for l, (x, xe), (y, ye) in content],
            fmt="none", ecolor="lightgrey"
        )
        # plot labels
        for label, (x, xe), (y, ye) in content:
            pylab.text(x, y, label, horizontalalignment='center', verticalalignment='center', size=8)
        pylab.title(title)

    def plot_baseline(content, baseline, title, feeds):
        """Plots offset versus baseline"""
        bl = []
        xx = []
        xxe = []
        lab = []
        col = []
        for l, (x, xe), (y, ye) in content:
            b = baseline.get(l, None)
            if b is None:
                print("baseline '{}' not found in MS ANTENNA table".format(l), file=log(0, "red"))
            else:
                lab += ["%s:%s" % (l, feeds[0]), "%s:%s" % (l, feeds[1])]
                col += ["blue", "red"]
                xx += [x, y]
                xxe += [xe, ye]
                bl += [b, b]
        pylab.axhline(1, color='lightgrey')
        pylab.errorbar(bl, xx, yerr=xxe, fmt="none", ecolor="lightgrey")
        # plot labels
        for l, x, y, c in zip(lab, bl, xx, col):
            pylab.text(x, y, l, horizontalalignment='center', verticalalignment='center', size=8, color=c)
        pylab.xlabel("Baseline, m.")
        pylab.title(title)

    def plot_hist(content, title):
        """Plots histogram"""
        values = [x for l, (x, xe), (y, ye) in content] + [y for l, (x, xe), (y, ye) in content]
        x0, x1 = min(values), max(values)
        if (x1-x0) > 1e-5:
            hist = pylab.hist(values)
            pylab.xlim(x0, x1)
        pylab.title(title)

    def plot_complex(content, title):
        """Plots x vs y"""
        # plot errors bars, if available
        pylab.axhline(0, color='lightgrey')
        pylab.axvline(1, color='lightgrey')
        pylab.errorbar(
            [x.real for l1, l2, (x, xe), (y, ye) in content], [x.imag for l1, l2, (x, xe), (y, ye) in content],
            [xe for l1, l2, (x, xe), (y, ye) in content], [xe for l1, l2, (x, xe), (y, ye) in content],
            fmt="none", ecolor="lightgrey"
        )
        pylab.errorbar(
            [y.real for l1, l2, (x, xe), (y, ye) in content], [y.imag for l1, l2, (x, xe), (y, ye) in content],
            [ye for l1, l2, (x, xe), (y, ye) in content], [ye for l1, l2, (x, xe), (y, ye) in content],
            fmt="none", ecolor="lightgrey"
        )
        # max plot amplitude -- max point plus 1/4th of the error bar
        maxa = max([max(_abs(x), _abs(y)) for l1, l2, (x, xe), (y, ye) in content])
        # plotlim = max([ abs(np.array([
        #                  getattr(v,attr)+sign*e/4 for v,e in (x,xe),(y,ye) for attr in 'real','imag' for sign in 1,-1
        #                ])).max()
        #   for l1,l2,(x,xe),(y,ye) in content ])
        minre, maxre, minim, maxim = 2, -2, 2, -2
        for l1, l2, (x, xe), (y, ye) in content:
            offs = np.array([getattr(v, attr) + sign * e / 4 for v, e in ((x, xe), (y, ye))
                                for attr in ('real', 'imag') for sign in (1, -1)])
            minre, maxre = min(x.real - xe / 4, y.real - ye / 4, minre), max(x.real + xe / 4, y.real + ye / 4, maxre)
            minim, maxim = min(x.imag - xe / 4, y.imag - ye / 4, minim), max(x.imag + xe / 4, y.imag + ye / 4, maxim)
        # plot labels
        for l1, l2, (x, xe), (y, ye) in content:
            pylab.text(x.real, x.imag, l1, horizontalalignment='center', verticalalignment='center', color='blue',
                       size=8)
            pylab.text(y.real, y.imag, l2, horizontalalignment='center', verticalalignment='center', color='red',
                       size=8)
        # pylab.xlim(-plotlim,plotlim)
        # pylab.ylim(-plotlim,plotlim)
        pylab.xlim(minre, maxre)
        pylab.ylim(minim, maxim)
        pylab.title(title + " (max %.5g)" % maxa)

    def plot_ants(content, title):
        """Plots x vs y"""
        # plot labels
        for p, gainlist in content:
            for label, color, (value, std) in gainlist:
                if value:
                    pylab.plot(p, value, 'w.')
                    pylab.text(p, value, label, horizontalalignment='center', verticalalignment='center', size=8,
                               color=color)
        pylab.xlim(-1, len(ms.antnames))
        make_antenna_xaxis(ms.antnames)
        pylab.title(title)

    nfreq, nant, _, _, _ = ig.shape
    # collect a list of valid RR/LL and RL/LR pairs (i.e. ones not all unity)

    # plot size and layout
    width, height = 8, 6  # size per sublot

    # loop over diag and offdiag elements
    components = [ (0, 0, 1, 1, "diag") ]
    if GD["bbc"]["compute-2x2"]:
        components.append((0, 1, 1, 0, "offdiag"))

    for i1,j1,i2,j2,label in components:
        # this is a 3x2 plot
        NR, NC = 3, 2
        # collect a list of valid RR/LL and RL/LR pairs (i.e. ones not all unity)
        valid_igs = []
        ifr_pairs = {}
        for p in range(nant):
            for q in range(p+1, nant):
                ifrname = ms.metadata.baseline_name[p,q]
                rr = ig[:, p, q, i1, j1]
                ll = ig[:, p, q, i2, j2]
                if not _is_unity(rr, ll):
                    valid_igs.append((ifrname, rr, ll))
                    ifr_pairs[ifrname] = p,q
        if not valid_igs:
            continue
        FEEDS = (ms.feeds[i1]+ms.feeds[j1]).upper(), (ms.feeds[i2]+ms.feeds[j2]).upper()

        norm_igs = [ (pq, _normifrgain(rr),_normifrgain(ll)) for pq, rr, ll in valid_igs ]
        complex_igs = [("{}:{}".format(pq, FEEDS[0]), "{}:{}".format(pq, FEEDS[1]),
                        _complexifrgain(rr), _complexifrgain(ll)) for pq, rr, ll in valid_igs]
        pylab.subplot(NR, NC, 3)
        plot_xy(norm_igs, "IFR gain amplitude (%s vs. %s)" % FEEDS)
        pylab.subplot(NR, NC, 4)
        plot_hist(norm_igs, "IFR gain histogram for %s and %s" % FEEDS)
        pylab.subplot(NR, NC, 1)
        plot_complex(complex_igs, "IFR complex %s %s gains" % FEEDS)
        igpa = {}
        igpa0 = {}
        igpa0_means = []
        for pq, rr, ll in valid_igs:
            p,q = ifr_pairs[pq]
            rr0 = _abs(rr - 1)
            ll0 = _abs(ll - 1)
            rr0[rr0 == 0] = ma.masked
            ll0[ll0 == 0] = ma.masked
            if not rr0.mask.all():
                igpa0_means += [rr0.mean()]
            if not ll0.mask.all():
                igpa0_means += [ll0.mean()]
            if not rr0.mask.all() or not ll0.mask.all():
                igpa0.setdefault(p, {})[q] = rr0, ll0
                igpa0.setdefault(q, {})[p] = rr0, ll0
            rr, ll = _normifrgain(rr), _normifrgain(ll)
            igpa.setdefault(p, []).append(("%s:%s" % (ms.antnames[q], FEEDS[0]), 'blue', rr))
            igpa.setdefault(q, []).append(("%s:%s" % (ms.antnames[p], FEEDS[0]), 'blue', rr))
            igpa.setdefault(p, []).append(("%s:%s" % (ms.antnames[q], FEEDS[1]), 'red', ll))
            igpa.setdefault(q, []).append(("%s:%s" % (ms.antnames[p], FEEDS[1]), 'red', ll))
        antennas = sorted(igpa.keys())
        pylab.subplot(NR, NC, 2)
        plot_ants([(p, igpa[p]) for p in antennas], "IFR %s %s gain amplitudes per antenna" % FEEDS)
        if baseline:
            pylab.subplot(NR, NC//2, 3)
            plot_baseline(norm_igs, baseline, "IFR gain amplitude vs. baseline length", FEEDS)
        save_figure("bbc-{}".format(label), width*NC, height*NR)

        # make per-antenna figure
        NC = 4
        NR = int(math.ceil(len(antennas) / float(NC)))
        offset = np.median(igpa0_means)
        for iant, pant in enumerate(antennas):
            pylab.subplot(NR, NC, iant + 1)
            igains = igpa0[pant]
            ants1 = sorted(igains.keys())
            for i, qant in enumerate(ants1):
                rr, ll = igains[qant]
                if rr.count() > 1:
                    a1, a2 = np.ma.flatnotmasked_edges(rr)
                    line, = pylab.plot(rr + i * offset, '-')
                    pylab.text(a1, rr[a1] + i * offset, "%s:%s" % (ms.antnames[qant], feeds[0]), horizontalalignment='left',
                               verticalalignment='center', size=8,
                               color=line.get_color())
                if ll.count() > 1:
                    a1, a2 = np.ma.flatnotmasked_edges(ll)
                    line, = pylab.plot(ll + i * offset, '-')
                    pylab.text(a2, ll[a2] + i * offset, "%s:%s" % (ms.antnames[qant], feeds[3]), horizontalalignment='right',
                               verticalalignment='center', size=8,
                               color=line.get_color())
            pylab.title("antenna %s" % ms.antnames[pant])
        save_figure("bbc-{}-ant".format(label), 8 * NC, 6 * NR)




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Handles the interface between measurement sets, CubiCal and Montblanc.
"""
from __future__ import print_function
from builtins import range
import numpy as np
import montblanc
import montblanc.util as mbu
import montblanc.impl.rime.tensorflow.ms.ms_manager as MS

from montblanc.impl.rime.tensorflow.sources import SourceProvider
from montblanc.impl.rime.tensorflow.sinks import SinkProvider

import datetime as dt
import pyrap.quanta as pq

from cubical.tools import logger, ModColor
log = logger.getLogger("MBSourceProvider")

class MSSourceProvider(SourceProvider):
    """
    Handles interface between CubiCal tiles and Montblanc simulation.
    """

    def __init__(self, tile, time_col, antea, anteb, ddid_col, uvw, freqs, sort_ind, nrows, do_pa_rotation=True):
        """
        Initialises this source provider.

        Args:
            tile (:obj:`~cubical.data_handler.Tile`):
                Tile object containing information about current data selection.
            uvw (np.darray):
                (n_row, 3) array of UVW coordinates.
            freqs (np.darray):
                (n_ddid,n_chan) array of frequencies.
            sort_ind (np.ndarray):
                Indices which will produce sorted data. Montblanc expects data to be ordered in a specific way.
            nrows (int):
                Number of rows in the UNPADDED data. This is necessary for the revised uvw code.

        """

        self._ms = tile.dh.ms
        self._ms_name = tile.dh.ms_name
        self._name = "Measurement set '{ms}'".format(ms=self._ms_name)

        self._antpos = tile.dh.antpos
        self._phadir = tile.dh.phadir

        self._ntime = len(np.unique(time_col))
        self._nchan = freqs.shape[1]
        self._nants = tile.dh.nants
        self._ncorr = tile.dh.ncorr
        self._nbl   = (self._nants*(self._nants - 1))/2
        self._freqs = freqs
        self._times = time_col
        self._antea = antea
        self._anteb = anteb
        self._ddids = ddid_col
        self._nddid = freqs.shape[0]
        self._uvwco = uvw                  #  data['uvwco']
        self._nrows = nrows
        self.sort_ind = sort_ind
        self.do_pa_rotation = do_pa_rotation

    def name(self):
        """ Returns name of associated source provider. """

        return self._name

    def updated_dimensions(self):
        """ Inform Montblanc of the dimensions assosciated with this source provider. """

        return [('ntime', self._ntime),
                ('nbl', self._nbl),
                ('na', self._nants),
                ('nchan', self._nchan*self._nddid),
                ('nbands', self._nddid),
                ('npol', 4),
                ('npolchan', 4*self._nchan)]

    def frequency(self, context):
        """ Provides Montblanc with an array of frequencies. """
        return self._freqs.reshape(context.shape).astype(context.dtype)

    def uvw(self, context):
        """ Provides Montblanc with an array of uvw coordinates. """

        # Figure out our extents in the time dimension and our global antenna and baseline sizes.

        (t_low, t_high) = context.dim_extents('ntime')

        # Figure out chunks in time (may be repetitious, but needed an easy fix).

        _, counts = np.unique(self._times[:self._nrows], return_counts=True)

        chunks = np.asarray(counts)

        # Compute per antenna uvw coordinates. Data must be ordered by time.
        # Per antenna uvw coordinates fail on data where time!=time_centroid.

        ant_uvw = mbu.antenna_uvw(self._uvwco[:self._nrows],
                                  self._antea[:self._nrows],
                                  self._anteb[:self._nrows],
                                  chunks,
                                  self._nants,
                                  check_missing=False,
                                  check_decomposition=False,
                                  max_err=100)

        return ant_uvw[t_low:t_high, ...].astype(context.dtype)

    def antenna1(self, context):
        """ Provides Montblanc with an array of antenna1 values. """

        lrow, urow = MS.uvw_row_extents(context)
        antenna1 = self._antea[self.sort_ind][lrow:urow]

        return antenna1.reshape(context.shape).astype(context.dtype)

    def antenna2(self, context):
        """ Provides Montblanc with an array of antenna2 values. """

        lrow, urow = MS.uvw_row_extents(context)
        antenna2 = self._anteb[self.sort_ind][lrow:urow]

        return antenna2.reshape(context.shape).astype(context.dtype)

    def parallactic_angles(self, context):
        """ Provides Montblanc with an array of parallactic angles. """

        # Time and antenna extents
        (lt, ut), (la, ua) = context.dim_extents('ntime', 'na')
        if not self.do_pa_rotation:
            return np.zeros(context.shape, dtype=context.dtype)

        def __mjd2dt(utc_timestamp):
            """
            Converts array of UTC timestamps to list of datetime objects for human readable printing
            """
            return [dt.datetime.utcfromtimestamp(pq.quantity(t, "s").to_unix_time()) for t in utc_timestamp]
        utc_times = np.unique(self._times[self.sort_ind])[lt:ut]
        dt_start = __mjd2dt([np.min(utc_times)])[0].strftime('%Y/%m/%d %H:%M:%S')
        dt_end = __mjd2dt([np.max(utc_times)])[0].strftime('%Y/%m/%d %H:%M:%S')
        log(2).print("Computing parallactic angles for times between %s and %s UTC" % (dt_start, dt_end))
        return mbu.parallactic_angles(
                        np.unique(self._times[self.sort_ind])[lt:ut],
                        self._antpos[la:ua],
                        self._phadir).reshape(context.shape).astype(context.dtype)

    def feed_angles(self, context):
        """ Provides Montblanc with an array of feed angles. """

        (la, ua) = context.dim_extents('na')
        # TODO(osmirnov)
        # Please fill me in
        return np.zeros(ua-la, dtype=context.dtype)

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, etraceback):
        self.close()

    def __str__(self):
        return self.__class__.__name__


class ColumnSinkProvider(SinkProvider):
    """
    Handles Montblanc output and makes it consistent with the measurement set.
    """
    def __init__(self, dh, freqshape, model, sort_ind):
        """
        Initialises this sink provider.

        Args:
            dh (:obj:`~cubical.data_handler.MSDataHandler`):
                Data handler object.
            freqshape (tuple):
                Shape of frequency array, i.e. (nddid, nchan)
            model (np.ndarray):
                Array of model visibilities into which output will be written.
            sort_ind (np.ndarray):
                Indices which will produce sorted data. Montblanc expects adata to be ordered.

        """

        self._model = model
        self._ncorr = dh.ncorr
        self._name = "Measurement Set '{ms}'".format(ms=dh.ms_name)
        self._dir = 0
        self.sort_ind = sort_ind
        self._nddid, self._chan_per_ddid = freqshape

    def name(self):
        """ Returns name of associated sink provider. """

        return self._name

    def set_direction(self, idir):
        """Sets current direction being simulated.

        Args:
            idir (int):
                Direction number, from 0 to n_dir-1
        """
        self._dir = idir

    def model_vis(self, context):
        """ Tells Montblanc how to handle the model visibility output. """

        (lt, ut), (lbl, ubl), (lc, uc) = context.dim_extents('ntime', 'nbl', 'nchan')

        lower, upper = MS.row_extents(context, ("ntime", "nbl"))

        ntime, nbl, nchan = context.dim_global_size('ntime', 'nbl', 'nchan')
        rows_per_ddid = ntime*nbl

        if self._ncorr == 1:
            sel = 0
        elif self._ncorr == 2:
            sel = slice(None, None, 3)
        else:
            sel = slice(None)

        for ddid_ind in range(self._nddid):
            offset = ddid_ind*rows_per_ddid
            lr = lower + offset
            ur = upper + offset
            lc = ddid_ind*self._chan_per_ddid
            uc = (ddid_ind+1)*self._chan_per_ddid
            self._model[self._dir, 0, lr:ur, :, :] += \
                    context.data[:,:,lc:uc,sel].reshape(-1, self._chan_per_ddid, self._ncorr)

    def __str__(self):
        return self.__class__.__name__

_mb_slvr = None

def simulate(src_provs, snk_provs, polarisation_type, opts):
    """
    Convenience function which creates and executes a Montblanc solver for the given source and
    sink providers.

    Args:
        src_provs (list):
            List of :obj:`~montblanc.impl.rime.tensorflow.sources.SourceProvider` objects. See
            Montblanc's documentation.
        snk_provs (list):
            List of :obj:`~montblanc.impl.rime.tensorflow.sinks.SinkProvider` objects. See
            Montblanc's documentation.
        opts (dict):
            Montblanc simulation options (see [montblanc] section in DefaultParset.cfg).
    """

    global _mb_slvr

    if _mb_slvr is None:
        slvr_cfg = montblanc.rime_solver_cfg(
            mem_budget=opts["mem-budget"]*1024*1024,
            dtype=opts["dtype"],
            polarisation_type=polarisation_type,
            device_type=opts["device-type"])

        _mb_slvr = montblanc.rime_solver(slvr_cfg)

    _mb_slvr.solve(source_providers=src_provs, sink_providers=snk_provs)

import atexit

def _shutdown_mb_slvr():

    global _mb_slvr

    if _mb_slvr is not None:
        _mb_slvr.close()

atexit.register(_shutdown_mb_slvr)




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Defines database interface
"""
from six import add_metaclass
import abc


@add_metaclass(abc.ABCMeta)
class iface_database:
    @abc.abstractmethod
    def __init__(self):
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def _create(self, filename, metadata={}, backup=True, **kw):
        """
        Creates a parameter database given by the filename and opens it in "create" mode.

        Args:
            filename (str): 
                Name of database.
            metadata (dict, optional): 
                Optional metadata to be stored in DB.
            backup (bool, optional):
                If True, and an old database with the same filename exists, make a backup.
            kw (dict):
                Keyword arguments.
        """
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def define_param(self, *args, **kw):
        """
        Defines a parameter. Only valid in "create" mode.

        Args:
            args (tuple):
                Positional arguments.
            kw (dict):
                Keyword arguments.
        """
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def add_chunk(self, name, array, grid={}):
        """
        Adds a slice of values for a parameter.

        Args:
            name (str):
                The parameter name e.g. "G".
            array (:obj:`~numpy.ma.core.MaskedArray`):
                The values which are to be added.
            grid (dict, optional):
                Grid coordinates for each sliced parameter axis. 

        """
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def close(self):
        """ Closes the database. """
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def _save_desc(self):
        """ Helper function. Writes accumulated parameter descriptions to filename.desc. """
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def _backup_and_rename(self, backup):
        """
        May create a backup and renames the temporary DB.

        Args:
            backup (bool):
                If True, creates a backup, otherwise just renames.
        """
        raise NotImplementedError("To be defined")
        
    @abc.abstractmethod
    def save(self, filename=None, backup=True):
        """
        Save the database.

        Args:
            filename (str, optional):
                Name of output file.
            backup (bool, optional):
                If True, create a backup.
        """
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def _load(self, filename):
        """
        Loads database from file. This will create arrays corresponding to the stored parameter
        shapes.

        Args:
            filename (str):
                Name of file to load.
        """
        raise NotImplementedError("To be defined")
    
    @abc.abstractmethod
    def names(self):
        """ Returns names of all defined parameters. """
        raise NotImplementedError("To be defined")

    @abc.abstractmethod
    def __contains__(self, name):
        raise NotImplementedError("To be defined")

    @abc.abstractmethod
    def __getitem__(self, name):
        raise NotImplementedError("To be defined")

    @abc.abstractmethod
    def get(self, name):
        """
        Gets Parameter object associated with the named parameter.

        Args:
            name (str):
                Name of parameter.

        Returns:
            :obj:`~cubical.param_db.Parameter`:
                The requested Parameter object.
        """
        raise NotImplementedError("To be defined")


from __future__ import print_function
from psutil import virtual_memory
from cubical import workers

from cubical.tools import logger, ModColor
log = logger.getLogger("wisdom")


def estimate_mem(data_handler, tile_list, data_opts, dist_opts):

    # Grab current memory information.

    mem = virtual_memory()

    tot_sys_mem = mem.total/(1024**3)  # Convert to value in GiB.

    # Grab an example tile - for now we assume that it is a representative
    # example. In principle we could loop over all tiles and look for the
    # absolute peak memeory usage. We do some fiddling to get the number of
    # chunks along each axis.

    example_tile = tile_list[0]
    chunks_per_tile = example_tile.total_tf_chunks()
    along_freq = max(len(cf) for cf in data_handler.freqchunks.values())
    along_time = chunks_per_tile//along_freq

    # Figure out the chunk dimensions. The ors handle the 0 case. TODO: The
    # time behaviour might not be safe.

    t_per_chunk = example_tile.rowchunks[0].timeslice.stop
    f_per_chunk = data_opts["freq-chunk"] or data_handler.all_freqs.size

    # Grab some necessary dimensions. n_dir may be zero when transferring.
    # The or catches this case.

    n_dir = len(data_handler.model_directions) or 1
    n_ant = data_handler.nants
    n_bl = (n_ant*(n_ant - 1))/2
    n_corr = example_tile.ncorr

    # Figure out the data points per tile. This is done assuming that the data
    # is still in rowlike format at this point. TODO: Verify behaviour with
    # multiple directions.

    data_points_per_tile = \
        n_dir*(along_time*t_per_chunk*n_bl)*(along_freq*f_per_chunk)*n_corr

    # Figure out the data points per chunk. This is done assuming that the
    # data is now going to be in the internal data representation. TODO:
    # verify that this works correctly when varying n_dir and n_ant.

    data_points_per_chunk = \
        n_dir*t_per_chunk*f_per_chunk*n_ant*n_ant*n_corr

    # Empirically determined values for the slope and intercept for each
    # solver process.

    w_slope, w_intercept = 5.77281225e-08, 4.53587065e-01

    w_mem_est = w_slope*data_points_per_chunk + w_intercept

    # Empirically determined values for the slope and intercept for the
    # I/O process.

    io_slope, io_intercept = 4.24736385e-08, 3.73958333e-01

    io_mem_est = io_slope*data_points_per_tile + io_intercept

    # Make a guess at the total memory use.

    tot_mem_est = w_mem_est*(workers.num_workers or 1) + io_mem_est

    # Log all of this to terminal.

    print("Detected a total of {:.2f}GiB of system memory.".format(
          tot_sys_mem), file=log)
    print("Per-solver (worker) memory use estimated at {:.2f}GiB: {:.2f}% of "
          "total system memory.".format(w_mem_est, 100*w_mem_est/tot_sys_mem),
          file=log)
    print("Peak I/O memory use estimated at {:.2f}GiB: {:.2f}% of total "
          "system memory.".format(io_mem_est, 100*io_mem_est/tot_sys_mem),
          file=log)
    print("Total peak memory usage estimated at {:.2f}GiB: {:.2f}% of total "
          "system memory.".format(tot_mem_est, 100*tot_mem_est/tot_sys_mem),
          file=log)

    if tot_mem_est > tot_sys_mem*dist_opts["safe"] and dist_opts["safe"]:

        raise MemoryError(
            "Estimated memory usage exceeds allowed pecentage of system "
            "memory. Memory usage can be reduced by lowering the number of "
            "chunks, the dimensions of each chunk or the number of worker "
            "processes. This error can suppressed by setting --dist-safe to "
            "zero.")

    return
